#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Module containing validators for various parameters.
"""
import re
import wx
import psychopy.experiment.utils
from psychopy.tools import stringtools
from psychopy.localization import _translate
from . import experiment
from .localizedStrings import _localized
from pkg_resources import parse_version
from ...visual.textbox2.fontmanager import FontManager

fontMGR = FontManager()

if parse_version(wx.__version__) < parse_version('4.0.0a1'):
    _ValidatorBase = wx.PyValidator
else:
    _ValidatorBase = wx.Validator

# Symbolic constants representing the 'kind' of warning for instances of
# `ValidatorWarning`.
VALIDATOR_WARNING_NONE = 0
VALIDATOR_WARNING_NAME = 1
VALIDATOR_WARNING_SYNTAX = 2
VALIDATOR_WARNING_FONT_MISSING = 3
VALIDATOR_WARNING_COUNT = 4  # increment when adding more


class ValidatorWarning:
    """Class for validator warnings.

    These are used internally by the `WarningManager`, do not create instances
    of this class unless you know what you're doing with them.

    Parameters
    ----------
    parent : wx.Window or wx.Panel
        Dialog associate with this warning.
    control : wx.Window or wx.Panel
        Control associated with the validator which threw this warning.
    msg : str
        Message text associated with the warning to be displayed.
    kind : int
        Symbolic constant representing the type of warning. Values can be one of
        `VALIDATOR_WARNING_NONE`, `VALIDATOR_WARNING_NAME`,
        `VALIDATOR_WARNING_SYNTAX` or `VALIDATOR_WARNING_FONT_MISSING`.

    """
    __slots__ = [
        '_parent',
        '_control',
        '_msg',
        '_kind']

    def __init__(self, parent, control, msg="", kind=VALIDATOR_WARNING_NONE):
        self.parent = parent
        self.control = control
        self.msg = msg
        self.kind = kind

    @property
    def parent(self):
        """Dialog associate with this warning (`wx.Window` or similar)."""
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def control(self):
        """Control associated with the validator which threw this warning
        (`wx.Window` or similar).
        """
        return self._control

    @control.setter
    def control(self, value):
        self._control = value

    @property
    def msg(self):
        """Message text associated with the warning to be displayed (`str`).
        """
        return self._msg

    @msg.setter
    def msg(self, value):
        self._msg = str(value)

    @property
    def kind(self):
        """Symbolic constant representing the type of warning (`int`). Values
        can be one of `VALIDATOR_WARNING_NONE`, `VALIDATOR_WARNING_NAME`,
        `VALIDATOR_WARNING_SYNTAX` or `VALIDATOR_WARNING_FONT_MISSING`.
        """
        return self._kind

    @kind.setter
    def kind(self, value):
        value = int(value)
        assert VALIDATOR_WARNING_NONE <= value < VALIDATOR_WARNING_COUNT
        self._kind = value

    @property
    def isSyntaxWarning(self):
        """`True` if this is a syntax warning (`bool`)."""
        return self._kind == VALIDATOR_WARNING_SYNTAX

    @property
    def isNameWarning(self):
        """`True` if this is a namespace warning (`bool`)."""
        return self._kind == VALIDATOR_WARNING_NAME

    @property
    def allowed(self):
        """`True` if this is a non-critical message which doesn't disable the OK button"""
        return self.kind in [VALIDATOR_WARNING_FONT_MISSING]


class WarningManager:
    """Manager for warnings produced by validators associated with controls
    within the component properties dialog. Assumes that the `parent` dialog
    uses a standardized convention for attribute names for all components.

    Each control can only have a single warning at a time in the current
    implementation of this class.

    Warnings
    --------
    Do not make strong references to instances of this class outside of the
    `parent` dialog. This class must be destroyed along with the `parent` object
    when the `parent` is deleted. This also goes for any `ValidatorWarning`
    objects referenced by this class.

    Parameters
    ----------
    parent : wx.Window
        Component properties dialog or panel.

    """
    def __init__(self, parent):
        self._parent = parent
        # Dictionary for storing warnings, keys are IDs for the controls that
        # produced them. In the future we should use wx object names to
        # reference these objects instead of IDs.
        self._warnings = {}

        # create an output label box for the parent
        self.output = wx.StaticText(
            self._parent, label="", style=wx.ALIGN_CENTRE_HORIZONTAL)
        self.output.SetForegroundColour(wx.RED)

    @property
    def OK(self):
        """`True` if there are no warnings (`bool`)."""
        if len(self._warnings) == 0:
            return True
        else:
            return all(warning.allowed for warning in self._warnings.values())

    @property
    def parent(self):
        """Parent dialog (`wx.Panel` or `wx.Window`). This attribute is
        read-only."""
        return self._parent

    @property
    def warnings(self):
        """Raw dictionary of warning objects (`dict`). Keys are IDs for the
        objects representing controls as integers, values are the
        `ValidatorWarning` instances associated with them."""
        return self._parent

    @property
    def messages(self):
        """List of warning messages (`list`). Messages are displayed in the
        order they have been added.
        """
        if not self._warnings:  # no warnings, return empty string
            return []

        warnings = self._warnings.values()  # warning objects
        return [warning.msg for warning in warnings]

    def getControlsWithWarnings(self):
        """Get a list of controls which have active warnings (`list`). You can
        use this to process controls which still have warnings registered to
        them.
        """
        if not self._warnings:  # no active warnings
            return []

        _, warnings = self._warnings.items()
        return [warning.control for warning in warnings]

    def setWarning(self, control, msg='', kind=VALIDATOR_WARNING_NONE):
        """Set a warning for a control. A control can only have one active
        warning associate with it at any given time.

        Parameters
        ----------
        control : wx.Window or wx.Panel
            Control to set an active warning for.
        msg : str
            Warning message text (e.g., "Syntax error").
        kind : int
            Symbolic constant representing the type of warning (e.g.,
            `VALIDATOR_WARN_SYNTAX`).

        """
        self._warnings[id(control)] = ValidatorWarning(
            self.parent, control, msg, kind)

    def getWarning(self, control):
        """Get an active warning associated with the control.

        Parameters
        ----------
        control : wx.Window or wx.Panel
            Control to check if there is a warning active against it.

        Returns
        -------
        ValidatorWarning or None
            Warning validator if there is warning, else None.

        """
        try:
            return self._warnings[id(control)]
        except KeyError:
            return None

    def clearWarning(self, control):
        """Clear the warning associated with a given control.

        Parameters
        ----------
        control : wx.Window or wx.Panel
            Control to clear any warnings against it.

        Returns
        -------
        bool
            `True` if the warning was cleared. `False` if there was no warning
            associated with the `control` provided.

        """
        wasCleared = True
        try:
            del self._warnings[id(control)]
        except KeyError:
            wasCleared = False

        return wasCleared

    def validate(self, control=None):
        """Validate one or many controls.

        Calling this will re-run validation on all controls which have active
        warnings presently registered to them. If the specified control(s) no
        longer generate warnings, it will be removed from the manager.

        This can be called make sure that all warnings have been addressed.

        """
        pass

    def _lockout(self, enable=True):
        """Lockout the dialog, preventing user changes from being applied.

        Parameters
        ----------
        enable : bool
            Lockout the dialog if `True`. `False` will re-enable the OK button.
            Assumes the parent dialog has an `ok` attribute which points to a
            `wx.Button` object or similar.

        """
        if hasattr(self.parent, 'ok'):
            okButton = self.parent.ok
        elif hasattr(self.parent, 'OKbtn'):  # another option ...
            okButton = self.parent.OKbtn
        else:
            # raise AttributeError("Parent object does not have an OK button.")
            return  # nop better here?

        if isinstance(okButton, wx.Button):
            okButton.Enable(enable)

    def showWarning(self):
        """Show the active warnings. Disables the OK button if present.
        """
        self._lockout(self.OK)  # enable / disable ok button

        # If there's any errors to show, show them
        messages = self.messages

        if messages:
            self.output.SetLabel("\n".join(messages))
        else:
            self.output.SetLabel("")

        # Update sizer
        sizer = self.output.GetContainingSizer()
        if sizer:
            sizer.Layout()


class BaseValidator(_ValidatorBase):
    """Component name validator for _BaseParamsDlg class. It depends on access
    to an experiment namespace.

    Validate calls check, which needs to be implemented per class.

    Messages are passed to user as text in nameOklabel.

    See Also
    --------
    _BaseParamsDlg

    """
    def __init__(self):
        super(BaseValidator, self).__init__()

    def Clone(self):
        return self.__class__()

    def Validate(self, parent):
        # we need to find the dialog to which the Validate event belongs
        # (the event might be fired by a sub-panel and won't have builder exp)
        while not hasattr(parent, 'warnings'):
            try:
                parent = parent.GetParent()
            except Exception:
                raise AttributeError("Could not find warnings manager")
        self.check(parent)

        return True

    def TransferFromWindow(self):
        return True

    def TransferToWindow(self):
        return True

    def check(self, parent):
        raise NotImplementedError


class NameValidator(BaseValidator):
    """Validation checks if the value in Name field is a valid Python
    identifier and if it does not clash with existing names.

    """
    def __init__(self):
        super(NameValidator, self).__init__()

    def check(self, parent):
        """Checks namespace.

        Parameters
        ----------
        parent : object
            Component properties dialog or similar.

        """
        control = self.GetWindow()
        newName = control.GetValue()
        msg, OK = '', True  # until we find otherwise
        if newName == '':
            msg = _translate("Missing name")
            OK = False
        else:
            namespace = parent.frame.exp.namespace
            used = namespace.exists(newName)
            sameAsOldName = bool(newName == parent.params['name'].val)
            if used and not sameAsOldName:
                # NOTE: formatted string literal doesn't work with _translate().
                # So, we have to call format() after _translate() is applied.
                msg = _translate(
                    "That name is in use (by {used}). Try another name."
                    ).format(used=_translate(used))
                OK = False
            elif not namespace.isValid(newName):  # valid as a var name
                msg = _translate("Name can only contain letters, numbers and underscores (_), no spaces or other symbols")
                OK = False
            # warn but allow, chances are good that its actually ok
            elif namespace.isPossiblyDerivable(newName):
                msg = _translate(namespace.isPossiblyDerivable(newName))
                OK = True

        if not OK:
            parent.warnings.setWarning(
                control, msg=msg, kind=VALIDATOR_WARNING_NAME)
        else:
            parent.warnings.clearWarning(control)

        parent.warnings.showWarning()


class CodeSnippetValidator(BaseValidator):
    """Validation checks if field value is intended to be Python code.
    If so, check that it is valid Python, and if valid, check whether
    it contains bad identifiers (currently only self-reference).

    @see: _BaseParamsDlg
    """

    def __init__(self, fieldName):
        super(CodeSnippetValidator, self).__init__()
        self.fieldName = fieldName
        try:
            self.displayName = _localized[fieldName]
        except KeyError:
            # should have all _localized[fieldName] from .localizedStrings
            # might as well try to do something useful if fail:
            self.displayName = _translate(fieldName)

    def Clone(self):
        return self.__class__(self.fieldName)

    def check(self, parent):
        """Checks python syntax of code snippets, and for self-reference.

        Note: code snippets simply use existing names in the namespace, like
        from condition-file headers. They do not add to the namespace (unlike
        Name fields). Code snippets in param fields will often be user-defined
        vars, especially condition names. Can also be expressions like
        random(1,100). Hard to know what will be problematic. But its always the
        case that self-reference is wrong.

        Parameters
        ----------
        parent : object
            Component properties dialog or similar.

        """
        # first check if there's anything to validate (and return if not)
        def _checkParamUpdates(parent):
            """Checks whether param allows updates. Returns bool."""
            if parent.params[self.fieldName].allowedUpdates is not None:
                # Check for new set with elements common to lists compared -
                # True if any elements are common
                return bool(
                    set(parent.params[self.fieldName].allowedUpdates) &
                    set(allowedUpdates))

        def _highlightParamVal(parent, error=False):
            """Highlights text containing error - defaults to black"""
            try:
                if error:
                    parent.paramCtrls[
                        self.fieldName].valueCtrl.SetForegroundColour("Red")
                else:
                    parent.paramCtrls[
                        self.fieldName].valueCtrl.SetForegroundColour("Black")
            except KeyError:
                pass

        # Get attributes of value control
        control = self.GetWindow()
        if not hasattr(control, 'GetValue'):
            return '', True  # mdc - why return anything here?

        val = control.GetValue()  # same as parent.params[self.fieldName].val
        if not isinstance(val, str):
            return '', True

        allowedUpdates = ['set every repeat', 'set every frame']
        # Set initials
        msg, OK = '', True  # until we find otherwise
        _highlightParamVal(parent)
        # What valType should code be treated as?
        codeWanted = psychopy.experiment.utils.unescapedDollarSign_re.search(val)
        isCodeField = bool(parent.params[self.fieldName].valType == 'code')

        # Validate as list
        # allKeyBoardKeys = list(key._key_names.values()) + [str(num) for num in range(10)]
        # allKeyBoardKeys = [key.lower() for key in allKeyBoardKeys]

        # Check if it is a Google font
        if self.fieldName == 'font' and not val.startswith('$'):
            fontInfo = fontMGR.getFontNamesSimilar(val)
            if not fontInfo:
                msg = _translate(
                    "Font `{val}` not found locally, will attempt to retrieve "
                    "from Google Fonts when this experiment next runs"
                ).format(val=val)
                parent.warnings.setWarning(
                    control, msg=msg, kind=VALIDATOR_WARNING_FONT_MISSING)
            else:
                parent.warnings.clearWarning(control)

        # Validate as code
        if codeWanted or isCodeField:
            # get var names from val, check against namespace:
            code = experiment.getCodeFromParamStr(val, target="PsychoPy")
            try:
                names = list(stringtools.getVariables(code))
                parent.warnings.clearWarning(control)
            except (SyntaxError, TypeError) as e:
                # empty '' compiles to a syntax error, ignore
                if not code.strip() == '':
                    _highlightParamVal(parent, True)
                    msg = _translate('Python syntax error in field `{}`:  {}')
                    msg = msg.format(self.displayName, code)
                    parent.warnings.setWarning(
                        control, msg=msg, kind=VALIDATOR_WARNING_SYNTAX)
            else:
                # Check whether variable param entered as a constant
                if isCodeField and _checkParamUpdates(parent):
                    if parent.paramCtrls[self.fieldName].getUpdates() not in allowedUpdates:
                        try:
                            eval(code)  # security risk here?
                        except NameError as e:
                            _highlightParamVal(parent, True)
                            # NOTE: formatted string literal doesn't work with _translate().
                            # So, we have to call format() after _translate() is applied.
                            msg = _translate(
                                "Looks like your variable '{code}' in "
                                "'{displayName}' should be set to update."
                                ).format(code=code, displayName=self.displayName)
                            parent.warnings.setWarning(
                                control, msg=msg, kind=VALIDATOR_WARNING_NAME)
                        except SyntaxError as e:
                            parent.warnings.setWarning(
                                control, msg=msg, kind=VALIDATOR_WARNING_SYNTAX)

                # namespace = parent.frame.exp.namespace
                # parent.params['name'].val is not in namespace for new params
                # and is not fixed as .val until dialog closes. Use getvalue()
                # to handle dynamic changes to Name field:
                if 'name' in parent.paramCtrls:  # some components don't have names
                    parentName = parent.paramCtrls['name'].getValue()
                    for name in names:
                        # `name` means a var name within a compiled code snippet
                        if name == parentName:
                            _highlightParamVal(parent, True)
                            msg = _translate(
                                'Python var `{}` in `{}` is same as Name')
                            msg = msg.format(name, self.displayName)
                            parent.warnings.setWarning(
                                control, msg=msg, kind=VALIDATOR_WARNING_NAME)
                    else:
                        parent.warnings.clearWarning(control)

                    for newName in names:
                        namespace = parent.frame.exp.namespace
                        if newName in [*namespace.user, *namespace.builder,
                                       *namespace.constants]:
                            # Continue if name is a variable
                            continue
                        if newName in [*namespace.nonUserBuilder, *namespace.numpy] \
                                and not re.search(newName+r"(?!\(\))", val):
                            # Continue if name is an external function being called correctly
                            continue
                        used = namespace.exists(newName)
                        sameAsOldName = bool(newName == parent.params['name'].val)
                        if used and not sameAsOldName:
                            # NOTE: formatted string literal doesn't work with _translate().
                            # So, we have to call format() after _translate() is applied.
                            msg = _translate(
                                "Variable name ${newName} is in use (by "
                                "{used}). Try another name."
                                ).format(newName=newName, used=_translate(used))
                            parent.warnings.setWarning(
                                control, msg=msg, kind=VALIDATOR_WARNING_NAME)
                        else:
                            parent.warnings.clearWarning(control)
        else:
            parent.warnings.clearWarning(control)

        parent.warnings.showWarning()  # show most recent warnings


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of using Cedrus RB730 hardware
"""

from psychopy.hardware import cedrus
from psychopy import core
import sys

rb730 = cedrus.RB730(7, baudrate=115200)

# Get RB info
print(rb730.getInfo())
# this is the time taken to send a signal to the unit and back via USB:
print(('roundTrip:', rb730.measureRoundTrip()))

core.wait(0.1)  # give chance to clear prev commands
rb730.resetBaseTimer()
rb730.resetTrialTimer()

# Test keys
print('push some keys (1 exits)')
sys.stdout.flush()
notAbort = True
while notAbort:
    keyEvents = rb730.waitKeyEvents(downOnly=False)
    for evt in keyEvents:
        print("key=%s, rt=%.4f, up/down=%i" % (evt.key, evt.rt, evt.direction))
        if evt.key == 1:
            notAbort = False

print('done')
print(('baseTime:', rb730.getBaseTimer()))
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Classes and functions for the script output."""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import re
import locale
import wx
import wx.richtext
import webbrowser
from psychopy.localization import _translate
from psychopy.alerts._alerts import AlertEntry
from psychopy.app.themes import ThemeMixin

_prefEncoding = locale.getpreferredencoding()


class ScriptOutputPanel(wx.richtext.RichTextCtrl, ThemeMixin):
    """Class for the script output window in Coder.

    Parameters
    ----------
    parent : :class:`wx.Window`
        Window this object belongs to.
    style : int
        Symbolic constants for style flags.
    size : ArrayLike or None
        Size of the control in pixels `(w, h)`. Use `None` for default.
    font : str or None
        Font to use for output, fixed-width is preferred. If `None`, the theme
        defaults will be used.
    fontSize : int or None
        Point size of the font. If `None`, the theme defaults will be used.

    """
    def __init__(self,
                 parent,
                 style=wx.TE_READONLY | wx.TE_MULTILINE | wx.BORDER_NONE,
                 size=None,
                 font=None,
                 fontSize=None):

        wx.richtext.RichTextCtrl.__init__(
            self,
            parent,
            id=wx.ID_ANY,
            value="",
            pos=wx.DefaultPosition,
            size=wx.DefaultSize if size is None else size,
            style=style,
            validator=wx.DefaultValidator,
            name=wx.TextCtrlNameStr)

        self.parent = parent
        self._font = font
        self._fontSize = fontSize
        self.Bind(wx.EVT_TEXT_URL, self.onURL)

    def write(self, inStr):
        """Write (append) text to the control.

        Formatting is automatically applied to the text assuming the text
        follows PsychoPy's standard formatting conventions.

        Parameters
        ----------
        inStr : str
            Text to append.

        """
        # mostly taken from the existing StdOutRich class used by runner
        self.MoveEnd()  # always 'append' text rather than 'writing' it

        if isinstance(inStr, AlertEntry):
            alert = inStr
            # Write Code
            self.BeginBold()
            self.BeginTextColour(wx.BLUE)
            self.BeginURL(alert.url)
            self.WriteText("Alert {}:".format(alert.code))
            self.EndURL()
            self.EndBold()
            self.EndTextColour()

            # Write Message
            self.BeginTextColour([0, 0, 0])
            self.WriteText(alert.msg)
            self.EndTextColour()

            # Write URL
            self.WriteText("\n\t"+_translate("For further info see "))
            self.BeginBold()
            self.BeginTextColour(wx.BLUE)
            self.BeginURL(alert.url)
            self.WriteText("{:<15}".format(alert.url))
            self.EndURL()
            self.EndBold()
            self.EndTextColour()

            self.Newline()
            self.ShowPosition(self.GetLastPosition())

            self._applyAppTheme()

            return

        # convert to unicode if needed
        if isinstance(inStr, bytes):
            try:
                inStr = inStr.decode('utf-8')
            except UnicodeDecodeError:
                inStr = inStr.decode(_prefEncoding)

        # process the line, apply formatting and append
        for thisLine in inStr.splitlines(True):
            try:
                thisLine = thisLine.replace("\t", "    ")
            except Exception as e:
                self.WriteText(str(e))

            if len(re.findall('".*", line.*', thisLine)) > 0:
                # this line contains a file/line location so write as URL
                # self.BeginStyle(self.urlStyle)  # this should be done with
                # styles, but they don't exist in wx as late as 2.8.4.0
                self.BeginBold()
                self.BeginTextColour(wx.BLUE)
                self.BeginURL(thisLine)
                self.WriteText(thisLine)
                self.EndURL()
                self.EndBold()
                self.EndTextColour()
            elif len(re.findall('WARNING', thisLine)) > 0:
                self.BeginTextColour([0, 150, 0])
                self.WriteText(thisLine)
                self.EndTextColour()
            elif len(re.findall('ERROR', thisLine)) > 0:
                self.BeginTextColour([150, 0, 0])
                self.WriteText(thisLine)
                self.EndTextColour()
            else:
                # line to write as simple text
                self.WriteText(thisLine)
        self._applyAppTheme()
        self.MoveEnd()  # go to end of stdout so user can see updated text
        self.ShowPosition(self.GetLastPosition())

    def onURL(self, evt):
        """Open link in default browser."""
        wx.BeginBusyCursor()
        try:
            if evt.String.startswith("http"):
                webbrowser.open(evt.String)
            else:
                # decompose the URL of a file and line number"""
                # "C:\Program Files\wxPython...\samples\hangman\hangman.py"
                filename = evt.GetString().split('"')[1]
                lineNumber = int(evt.GetString().split(',')[1][5:])
                self.GetTopLevelParent().gotoLine(filename, lineNumber)
        except Exception as e:
            print("##### Could not open URL: {} #####\n".format(evt.String))
            print(e)
        wx.EndBusyCursor()


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""All tests in this file involve rapidly changing colours, do not run these
tests in a setting where you can view the output if you have photosensitive
epilepsy.

"""
from psychopy.alerts._errorHandler import _BaseErrorHandler
from psychopy.tests import utils
from psychopy import visual, colors
import numpy as np

# Define expected values for different spaces
exemplars = [
    {'rgb': ( 1.00,  1.00,  1.00), 'rgb255': (255, 255, 255), 'hsv': (  0, 0.00, 1.00), 'hex': '#ffffff', 'named': 'white'},  # Pure white
    {'rgb': ( 0.00,  0.00,  0.00), 'rgb255': (128, 128, 128), 'hsv': (  0, 0.00, 0.50), 'hex': '#808080', 'named': 'gray'},  # Mid grey
    {'rgb': (-1.00, -1.00, -1.00), 'rgb255': (  0,   0,   0), 'hsv': (  0, 0.00, 0.00), 'hex': '#000000', 'named': 'black'},  # Pure black
    {'rgb': ( 1.00, -1.00, -1.00), 'rgb255': (255,   0,   0), 'hsv': (  0, 1.00, 1.00), 'hex': '#ff0000', 'named': 'red'},  # Pure red
    {'rgb': (-1.00,  1.00, -1.00), 'rgb255': (  0, 255,   0), 'hsv': (120, 1.00, 1.00), 'hex': '#00ff00', 'named': 'lime'},  # Pure green
    {'rgb': (-1.00, -1.00,  1.00), 'rgb255': (  0,   0, 255), 'hsv': (240, 1.00, 1.00), 'hex': '#0000ff', 'named': 'blue'},  # Pure blue
    # Psychopy colours
    {'rgb': (-0.20, -0.20, -0.14), 'rgb255': (102, 102, 110), 'hsv': (240, 0.07, 0.43), 'hex': '#66666e'},  # grey
    {'rgb': ( 0.35,  0.35,  0.38), 'rgb255': (172, 172, 176), 'hsv': (240, 0.02, 0.69), 'hex': '#acacb0'},  # light grey
    {'rgb': ( 0.90,  0.90,  0.90), 'rgb255': (242, 242, 242), 'hsv': (  0, 0.00, 0.95), 'hex': '#f2f2f2'},  # offwhite
    {'rgb': ( 0.90, -0.34, -0.29), 'rgb255': (242,  84,  91), 'hsv': (357, 0.65, 0.95), 'hex': '#f2545b'},  # red
    {'rgb': (-0.98,  0.33,  0.84), 'rgb255': (  2, 169, 234), 'hsv': (197, 0.99, 0.92), 'hex': '#02a9ea'},  # blue
    {'rgb': (-0.15,  0.60, -0.09), 'rgb255': (108, 204, 116), 'hsv': (125, 0.47, 0.80), 'hex': '#6ccc74'},  # green
    {'rgb': ( 0.85,  0.18, -0.98), 'rgb255': (236, 151,   3), 'hsv': ( 38, 0.99, 0.93), 'hex': '#ec9703'},  # orange
    {'rgb': ( 0.89,  0.65, -0.98), 'rgb255': (241, 211,   2), 'hsv': ( 52, 0.99, 0.95), 'hex': '#f1d302'},  # yellow
    {'rgb': ( 0.53,  0.49,  0.94), 'rgb255': (195, 190, 247), 'hsv': (245, 0.23, 0.97), 'hex': '#c3bef7'},  # violet
]
# A few values which are likely to mess things up
tykes = [
    {'rgba': ( 1.00,  1.00,  1.00, 0.50), 'rgba255': (255, 255, 255, 0.50), 'hsva': (  0, 0.00, 1.00, 0.50)},  # Make sure opacities work in every space
    {'rgba': "white", 'rgba255': "white", "hsva": "white", "hex": "white", "rgb255": "#ffffff"},  # Overriding colorSpace with hex or named values
    {'rgba': None, 'named': None, 'hex': None, 'hsva': None},  # None as a value
]

class Test_Window:
    """Some tests just for the window - we don't really care about what's drawn inside it
    """
    @classmethod
    def setup_class(self):
        self.win = visual.Window([128,128], pos=[50,50], allowGUI=False, autoLog=False)
        self.error = _BaseErrorHandler()

    @classmethod
    def teardown_class(self):
        self.win.close()

    # Begin test
    def test_colors(self):
        for colorSet in exemplars + tykes:
            # Construct matrix of space pairs
            spaceMatrix = []
            for space1 in colorSet:
                spaceMatrix.extend([[space1, space2] for space2 in colorSet if space2 != space1])
            # Compare each space pair for consistency
            for space1, space2 in spaceMatrix:
                col1 = colors.Color(colorSet[space1], space1)
                col2 = colors.Color(colorSet[space2], space2)
                closeEnough = all(abs(col1.rgba[i]-col2.rgba[i])<0.02 for i in range(4))
                # Check that valid color has been created
                assert (bool(col1) and bool(col2))
                # Check setters
                assert (col1 == col2 or closeEnough)

    def test_window_colors(self):
        # Iterate through color sets
        for colorSet in exemplars + tykes:
            for space in colorSet:
                # Set window color
                self.win.colorSpace = space
                self.win.color = colorSet[space]
                self.win.flip()
                utils.comparePixelColor(self.win, colors.Color(colorSet[space], space))

    def test_shape_colors(self):
        # Create rectangle with chunky border
        obj = visual.Rect(self.win, units="pix", pos=(0,0), size=(128, 128), lineWidth=10)
        # Iterate through color sets
        for colorSet in exemplars + tykes:
            for space in colorSet:
                # Check border color
                obj.colorSpace = space
                obj.borderColor = colorSet[space]
                obj.fillColor = 'white'
                obj.opacity = 1  # Fix opacity at full as this is not what we're testing
                self.win.flip()
                obj.draw()
                if colorSet[space]: # skip this comparison if color is None
                    utils.comparePixelColor(self.win, colors.Color(colorSet[space], space), coord=(1, 1))
                utils.comparePixelColor(self.win, colors.Color('white'), coord=(50, 50))
                # Check fill color
                obj.colorSpace = space
                obj.fillColor = colorSet[space]
                obj.borderColor = 'white'
                obj.opacity = 1  # Fix opacity at full as this is not what we're testing
                self.win.flip()
                obj.draw()
                if colorSet[space]: # skip this comparison if color is None
                    utils.comparePixelColor(self.win, colors.Color(colorSet[space], space), coord=(50, 50))
                utils.comparePixelColor(self.win, colors.Color('white'), coord=(1,1))

                # Testing foreColor is already done in test_textbox

    def test_element_array_colors(self):
        # Create element array with two elements covering the whole window in two block colours
        obj = visual.ElementArrayStim(self.win, units="pix",
                                      fieldPos=(0, 0), fieldSize=(128, 128), fieldShape='square', nElements=2,
                                      sizes=[[64, 128], [64, 128]], xys=[[-32, 0], [32, 0]], elementMask=None, elementTex=None)
        # Iterate through color sets
        for colorSet in exemplars + tykes:
            for space in colorSet:
                if space not in colors.strSpaces and not isinstance(colorSet[space], (str, type(None))):
                    # Check that setting color arrays renders correctly
                    obj.colorSpace = space
                    col1 = np.array(colorSet[space]).reshape((1, -1))
                    col2 = getattr(colors.Color('black'), space).reshape((1, -1))
                    obj.colors = np.append(col1, col2, 0)  # Set first color to current color set, second to black in same color space
                    obj.opacity = 1  # Fix opacity at full as this is not what we're testing
                    self.win.flip()
                    obj.draw()
                    utils.comparePixelColor(self.win, colors.Color(colorSet[space], space), coord=(10, 10))
                    utils.comparePixelColor(self.win, colors.Color('black'), coord=(10, 100))

    def test_visual_helper(self):
        # Create rectangle with chunky border
        obj = visual.Rect(self.win, units="pix", pos=(0, 0), size=(128, 128), lineWidth=10)
        # Iterate through color sets
        for colorSet in exemplars + tykes:
            for space in colorSet:
                # Check border color
                visual.helpers.setColor(obj,
                                        color=colorSet[space], colorSpace=space,
                                        colorAttrib="borderColor")
                obj.fillColor = 'white'
                obj.opacity = 1  # Fix opacity at full as this is not what we're testing
                self.win.flip()
                obj.draw()
                if colorSet[space]:  # skip this comparison if color is None
                    utils.comparePixelColor(self.win, colors.Color(colorSet[space], space), coord=(1, 1))
                utils.comparePixelColor(self.win, colors.Color('white'), coord=(50, 50))
                # Check fill color
                visual.helpers.setColor(obj,
                                        color=colorSet[space], colorSpace=space,
                                        colorAttrib="fillColor")
                obj.borderColor = 'white'
                obj.opacity = 1  # Fix opacity at full as this is not what we're testing
                self.win.flip()
                obj.draw()
                if colorSet[space]:  # skip this comparison if color is None
                    utils.comparePixelColor(self.win, colors.Color(colorSet[space], space), coord=(50, 50))
                utils.comparePixelColor(self.win, colors.Color('white'), coord=(1, 1))
        # Check color addition
        obj.fillColor = 'white'
        visual.helpers.setColor(obj,
                                color='black',
                                colorAttrib='fillColor',
                                operation='+')
        self.win.flip()
        obj.draw()
        utils.comparePixelColor(self.win, colors.Color('white') + colors.Color('black'), coord=(50, 50))
        # Check color subtraction
        obj.fillColor = 'grey'
        visual.helpers.setColor(obj,
                                color='black',
                                colorAttrib='fillColor',
                                operation='-')
        self.win.flip()
        obj.draw()
        utils.comparePixelColor(self.win, colors.Color('grey') - colors.Color('black'), coord=(50, 50))

        # Check alerts
        visual.helpers.setColor(obj, color="white", colorSpaceAttrib="fillColorSpace", rgbAttrib="fillRGB")
        assert any(err.code == 8105 for err in self.error.alerts), "Alert 8105 not triggered"
        assert any(err.code == 8110 for err in self.error.alerts), "Alert 8110 not triggered"

    def test_contrast(self):
        # Create rectangle with chunky border
        obj = visual.Rect(self.win, units="pix", pos=(0, 0), size=(128, 128), lineWidth=10)
        # Set its colors to be rgb extremes
        obj.fillColor = 'red'
        obj.borderColor = 'blue'
        obj.opacity = 1  # Fix opacity at full as this is not what we're testing
        # Halve contrast
        obj.contrast = 0.5
        # Refresh
        self.win.flip()
        obj.draw()
        # Check rendered color
        utils.comparePixelColor(self.win, colors.Color(( 0.5, -0.5, -0.5), "rgb"), coord=(50, 50))
        utils.comparePixelColor(self.win, colors.Color((-0.5, -0.5,  0.5), "rgb"), coord=(1, 1))


def test_color_operators():
    """Test for operators used to compare colors."""
    red255 = colors.Color((255, 0, 0), space='rgb255')
    redRGB = colors.Color((1, -1, -1), space='rgb')
    redRGB1 = colors.Color((1, 0, 0), space='rgb1')

    assert (red255 == redRGB == redRGB1)



# -*- coding: utf-8 -*-
"""
Tests timestamp accuracy and event delay of the keyboard event handling
backend specified by KB_BACKEND using a MilliKey device.

* This test / demo requires that a MilliKey device is connected to the computer being tested.

KB_BACKEND must be one of:

    - 'psychopy.event': Use PsychoPy's original pyglet based event handling backend.
    - 'psychopy.keyboard': Use keyboard.Keyboard class running PsychHID based backend.
    - 'psychopy.keyboard.iohub': Use keyboard.Keyboard class running iohub based backend.
    - 'psychopy.iohub': Use the psychopy.iohub event framewok directly.
    - 'psychhid': Use an alternative implementation using PsychHID, only for testing.
"""
import os
import string
from psychopy import core, visual, event
import serial
import numpy
import json
import time

# KB_BACKEND = 'psychopy.event'
KB_BACKEND = 'psychopy.keyboard'
# KB_BACKEND = 'psychopy.keyboard.iohub'
# KB_BACKEND = 'psychopy.iohub'
# KB_BACKEND = 'psychhid'

# Used by 'psychopy.keyboard' and 'psychopy.keyboard.iohub' backends
psychopy_keyboard_waitRelease = False

# Number of keyboard events to generate.
kgen_count = 10

# min, max msec duration for generated keypresses.
min_dur, max_dur = 50, 300

# Update the display each loop
# while waiting for event ...
update_display_while_waiting = True

# Sleep sleep_while_wait each loop
# while waiting for event.
# 0 == no call to sleep
sleep_while_wait = 0


#
# Helper functions
#


def getSerialPorts():
    available = []
    if os.name == 'nt':  # Windows
        for i in range(1, 256):
            try:
                sport = 'COM' + str(i)
                s = serial.Serial(sport, baudrate=128000)
                available.append(sport)
                s.close()
            except (serial.SerialException, ValueError):
                pass
    else:  # Mac / Linux
        from serial.tools import list_ports
        available = [port[0] for port in list_ports.comports()]
    return available


def getMilliKeyDevices():
    devices = []
    available = getSerialPorts()
    for p in available:
        try:
            mkey_sport = serial.Serial(p, baudrate=128000, timeout=1.0)
            while mkey_sport.readline():
                pass
            mkey_sport.write(b"GET CONFIG\n")
            rx_data = mkey_sport.readline()
            if rx_data:
                rx_data = rx_data[:-1].strip()
                try:
                    mkconf = json.loads(rx_data)
                    mkconf['port'] = p
                    devices.append(mkconf)
                except:
                    raise RuntimeError("ERROR: {}".format(rx_data))
            mkey_sport.close()
        except:
            pass
    return devices

#
# Run Test
#


if __name__ == "__main__":
    millikeys = getMilliKeyDevices()
    if len(millikeys) == 0:
        print("No Millikey device detected. Exiting test...")
        core.quit()
    
    mk_serial_port = millikeys[0]['port']
    # min, max msec MilliKey will wait before issuing the requested key press event.
    min_delay, max_delay = 1, 3
    possible_kgen_chars = string.ascii_lowercase + string.digits
    io = None
    ptb_kb = None
    getKeys = None

    # psychhid event timestamps are in psychtoolbox.GetSecs
    # so make test use that timebase.
    getTime = core.getTime
    if KB_BACKEND in ['psychhid', 'psychopy.keyboard']:
        from psychtoolbox import GetSecs
        getTime = GetSecs

    # psychopy.event is used as a secondary backend in case the primary backend
    # being tested (KB_BACKEND) fails to detect a keypress event.
    # Only used if KB_BACKEND != psychopy.event
    backup_getKeys = None
    if KB_BACKEND != 'psychopy.event':
        def backup_getKeys():
            return event.getKeys(timeStamped=True)

    # Create getKeys function to test desired keyboard backend     
    if KB_BACKEND == 'psychopy.event':
        def getKeys():
            return event.getKeys(timeStamped=True)
    elif KB_BACKEND == 'psychopy.keyboard':
        from psychopy.hardware.keyboard import Keyboard as ptbKeyboard
        ptb_kb = ptbKeyboard()
        def getKeys():
            keys = ptb_kb.getKeys(waitRelease=psychopy_keyboard_waitRelease)
            return [(k.name, k.tDown) for k in keys]
    elif KB_BACKEND == 'psychopy.iohub':
        from psychopy.iohub import launchHubServer
        io = launchHubServer()
        def getKeys():
            keys = io.devices.keyboard.getPresses()
            return [(k.key, k.time) for k in keys]
    elif KB_BACKEND == 'psychopy.keyboard.iohub':
        from psychopy.iohub import launchHubServer
        io = launchHubServer()
        from psychopy.hardware.keyboard import Keyboard as ptbKeyboard
        ptb_kb = ptbKeyboard()
        def getKeys():
            keys = ptb_kb.getKeys(waitRelease=psychopy_keyboard_waitRelease)
            return [(k.name, k.tDown) for k in keys]
    elif KB_BACKEND == 'psychhid':
        from psychtoolbox import PsychHID
        ptb_keys = [1] * 256
        PsychHID('KbQueueCreate', [], ptb_keys)
        PsychHID('KbQueueStart')
        
        def getKeys():
            keys = []
            while PsychHID('KbQueueFlush'):
                evt = PsychHID('KbQueueGetEvent')[0]
                if evt['Pressed']:
                    K = chr(int(evt['CookedKey'])).lower()
                    keys.append((K, evt['Time']))
            return keys
    
    def close_backend():
        if KB_BACKEND == 'psychhid':
            PsychHID('KbQueueStop')
        elif io:
            io.quit()

    win = visual.Window([800, 400])  #, fullscr=True, allowGUI=False)
    r = win.getMsPerFrame(60)

    print()
    print("Test Settings: ")
    print("\tEvent backend: ", KB_BACKEND)
    if KB_BACKEND.find('psychopy.keyboard') >= 0:
        print("\tpsychopy_keyboard_waitRelease: ", psychopy_keyboard_waitRelease)
    print("\tupdate_display_while_waiting: ", update_display_while_waiting)
    print("\tsleep_while_wait (sec): ", sleep_while_wait)
    print("\tmin_delay, max_delay (msec): ", (min_delay, max_delay))
    print("\tmin_dur, max_dur (msec): ", (min_dur, max_dur))
    print("\twin.getMsPerFrame(): ", r)
    print()

    ts_errors = numpy.zeros(kgen_count, dtype=numpy.float64)
    evt_delays = numpy.zeros(kgen_count, dtype=numpy.float64)

    mk_sconn = serial.Serial(mk_serial_port, baudrate=128000, timeout=0.01)
    

    txt1 = "Testing %s backend.\nMilliKey Generating Key Press:\n Key: [%s], Duration: %d msec.\n%d of %d events."
    msg = visual.TextStim(win, text=txt1)
    dotPatch = visual.DotStim(win, color=(0.0, 1.0, 0.0), dir=270,
                              nDots=223, fieldShape='sqr',
                              fieldPos=(0.0, 0.0), fieldSize=1.5,
                              dotLife=50, signalDots='same',
                              noiseDots='direction', speed=0.01,
                              coherence=0.9)
    
    def drawAndFlip(txt=None):
        if txt:
            msg.text = txt
        dotPatch.draw()
        msg.draw()
        return win.flip()
    
    drawAndFlip()
    print()
    count = 0
    no_event_count = 0
    while count < kgen_count:
        kchar = possible_kgen_chars[count % (len(possible_kgen_chars))]
        press_duration = int(numpy.random.randint(min_dur, max_dur))
        delay_evt_usec = int(numpy.random.randint(min_delay, max_delay))*1000
        evt_delay_sec = delay_evt_usec/1000.0/1000.0

        drawAndFlip(txt1 % (KB_BACKEND, kchar, press_duration, count+1, kgen_count))

        if ptb_kb:
            ptb_kb.clock.reset()

        # Instruct MilliKey device to:
        #  - generate a key press delay_evt_usec after receiving the KGEN command
        #  - generate key release event press_duration after press event is sent.
        kgen_cmd = "KGEN {} {} {}\n".format(kchar, press_duration, delay_evt_usec).encode()
        mk_sconn.write(kgen_cmd)
        mk_sconn.flush()
        # stime is the time the KGEN command was sent to the MilliKey device.
        # plus the event offset the device is using.
        stime = getTime()+evt_delay_sec

        # Keep checking for key press events until one is received
        kb_presses = getKeys()
        last_check_time = ctime = getTime()
        while not kb_presses:
            if getTime() - ctime > 2.0:
                # Report missed event.
                print("*WARNING: {} did not detect key press '{}' of duration {}.".format(KB_BACKEND, kchar, press_duration))
                if backup_getKeys:
                    presses = backup_getKeys()
                    if presses:
                        print("\tHowever, event.getKeys() received: ", presses)
                no_event_count += 1
                break
            if update_display_while_waiting:
                drawAndFlip()
            if sleep_while_wait > 0:
                time.sleep(sleep_while_wait)
            kb_presses = getKeys()
            last_check_time = getTime()

        # to clear backup getKeys()
        if backup_getKeys:
            _ = backup_getKeys()

        if kb_presses:
            if len(kb_presses) > 1:
                txt = "Error: {} keypress events detected at once: {}\n"
                txt += "Was a keyboard key or MilliKey button pressed during the test?"
                print(txt.format(len(kb_presses), kb_presses))
                close_backend()
                core.quit()
            
            # Get the key name and timestamp of event
            kpress, ktime = kb_presses[0]
        
            # Ensure we got the key we were expecting.....
            if kchar == kpress:
                ts_errors[count] = ktime - stime
                evt_delays[count] = last_check_time-stime
                count += 1
            else:
                txt = "Error: Keyboard Key != Key Press Issued ([{}] vs [{}]).\n"
                txt += "Was a keyboard key or MilliKey button pressed during the test?"
                print(txt.format(kchar, kpress))
                close_backend()
                core.quit()
        
            # Wait until after MilliKey has issued associated key release event.
            ctime = getTime()
            while getTime() - ctime < (press_duration/1000.0)*1.25:
                if update_display_while_waiting:
                    drawAndFlip()
                if sleep_while_wait > 0:
                    time.sleep(sleep_while_wait)
                getKeys()
    
    # Done test, close backend if needed
    close_backend()
    
    mk_sconn.close()
    win.close()
    
    # Print Results
    if no_event_count > 0:
        print()
        print("** WARNING: '%s' missed %d keypress events (%.6f percent)." % (KB_BACKEND, no_event_count,
                                                                              (no_event_count/count)*100))
        print()
    # Convert times to msec.
    evt_results = ts_errors[:count] * 1000.0
    print("%s Timestamp Accuracy Stats" % KB_BACKEND)
    print("\tCount: {}".format(evt_results.shape))
    print("\tAverage: {:.3f} msec".format(evt_results.mean()))
    print("\tMedian: {:.3f} msec".format(numpy.median(evt_results)))
    print("\tMin: {:.3f} msec".format(evt_results.min()))
    print("\tMax: {:.3f} msec".format(evt_results.max()))
    print("\tStdev: {:.3f} msec".format(evt_results.std()))
    print()
    # Convert times to msec.
    evt_results = evt_delays[:count] * 1000.0
    print("%s Keypress Event Delays" % KB_BACKEND)
    print("\tCount: {}".format(evt_results.shape))
    print("\tAverage: {:.3f} msec".format(evt_results.mean()))
    print("\tMedian: {:.3f} msec".format(numpy.median(evt_results)))
    print("\tMin: {:.3f} msec".format(evt_results.min()))
    print("\tMax: {:.3f} msec".format(evt_results.max()))
    print("\tStdev: {:.3f} msec".format(evt_results.std()))
    
    core.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Dialog prompting the user to configure their system for Linux specific
optimizations.
"""

__all__ = [
    'LinuxConfigDialog',
    'linuxConfigFileExists'
]

import os.path
import sys

import wx
from .ui import BaseLinuxConfigDialog

# Text that appears at the top of the dialog with provides instructions to the
# user.
_introStr = (
    u"For optimal performance on Linux, Psychtoolbox requires additional "
    u"configuration changes to be made to this system by entering the "
    u"following commands into your terminal:"
)

# config file path
_confPath = u"/etc/security/limits.d/99-psychopylimits.conf"

# these are the commands we want the user to run in their terminal
_cmdStr = (
    u"sudo groupadd --force psychopy\n\n"
    u"sudo usermod -a -G psychopy $USER\n\n"
    u"sudo gedit {fpath}\n"
    u"@psychopy - nice -20\n"
    u"@psychopy - rtprio 50\n"
    u"@psychopy - memlock unlimited")
_cmdStr = _cmdStr.format(fpath=_confPath)


class LinuxConfigDialog(BaseLinuxConfigDialog):
    """Class for the Linux one-time setup dialog.

    This dialog appears to users running PsychoPy on Linux for the first time.
    It prompts the user to open a terminal and enter the indicated commands that
    configure the environment to run Psychtoolbox with optimal settings.

    This dialog will appear at startup until the user completes the
    configuration steps. We determine if configuration is complete by testing if
    the expected file is present at the required location.

    Parameters
    ----------
    parent : wx.Window or None
        Dialog parent.
    timeout : int or None
        Milliseconds to keep the dialog open. Setting to `None` keeps the dialog
        open. Specify a time if we're running in a test environment to prevent
        the UI from locking up the test suite.

    """
    def __init__(self, parent, timeout=None):
        BaseLinuxConfigDialog.__init__(self, parent)

        # intro text which provides instructions for the user
        self.lblIntro.SetLabel(_introStr)
        self.lblIntro.Wrap(640)

        # make the text box show the commands for the user to enter into their
        # terminal
        self.txtCmdList.SetValue(_cmdStr)
        self.txtCmdList.SetMinSize((320, 240))

        # redo the layout
        self.DoLayoutAdaptation()

        self.timeout = timeout

        if self.timeout is not None:
            timeout = wx.CallLater(self.timeout, self.Close)
            timeout.Start()

    def OnOpenTerminal(self, event):
        event.Skip()

    def OnCopy(self, event):
        """Called when the copy button is clicked.
        """
        # check if we have a selection
        start, end = self.txtCmdList.GetSelection()
        if start != end:
            txt = self.txtCmdList.GetStringSelection()
        else:
            txt = self.txtCmdList.GetValue()

        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(wx.TextDataObject(txt))
            wx.TheClipboard.Close()

        event.Skip()

    def OnDone(self, event):
        """Called when the 'Done' button is clicked.
        """
        if not linuxConfigFileExists():
            msgText = (
                u"Setup does not appear to be complete, would you like to "
                u"continue anyways?\nNote that PTB audio timing performance "
                u"may not be optimal this session until this configuration "
                u"step is completed."
            )
            # show dialog
            dlg = wx.MessageDialog(
                self,
                msgText,
                style=wx.YES_NO | wx.YES_DEFAULT
            )
            result = dlg.ShowModal()
            if result == wx.ID_YES:
                self.Close()
            else:
                event.Skip()

            dlg.Destroy()


def linuxConfigFileExists():
    """Check if the required configuration file has been written.

    Returns
    -------
    bool
        `True` if the file exists or not on Linux.

    """
    # if not on linux, just pass it
    if sys.platform != 'linux':
        return True

    return os.path.isfile(_confPath)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to file and directory handling
"""
import os
import shutil
import sys
import atexit
import codecs
import numpy as np
import json
import json_tricks

try:
    import cPickle as pickle
except ImportError:
    import pickle

from psychopy import logging
from psychopy.tools.fileerrortools import handleFileCollision


def toFile(filename, data):
    """Save data (of any sort) as a pickle file.

    simple wrapper of the cPickle module in core python
    """
    f = open(filename, 'wb')
    pickle.dump(data, f)
    f.close()


def fromFile(filename, encoding='utf-8-sig'):
    """Load data from a psydat, pickle or JSON file.

    Parameters
    ----------
    encoding : str
        The encoding to use when reading a JSON file. This parameter will be
        ignored for any other file type.

    """
    filename = pathToString(filename)
    if filename.endswith('.psydat'):
        with open(filename, 'rb') as f:
            try:
                contents = pickle.load(f)
            except UnicodeDecodeError:
                f.seek(0)  # reset to start of file to try again
                contents = pickle.load(f, encoding='latin1')  # python 2 data files
            # if loading an experiment file make sure we don't save further
            # copies using __del__
            if hasattr(contents, 'abort'):
                contents.abort()
            return contents
    elif filename.endswith('pickle'):
        with open(filename, 'rb') as f:
            contents = pickle.load(f)
        return contents
    elif filename.endswith('.json'):
        with codecs.open(filename, 'r', encoding=encoding) as f:
            contents = json_tricks.load(f)

        # Restore RNG if we load a TrialHandler2 object.
        # We also need to remove the 'temporary' ._rng_state attribute that
        # was saved with it.
        from psychopy.data import TrialHandler2
        if isinstance(contents, TrialHandler2):
            contents._rng = np.random.default_rng()
            contents._rng.bit_generator.state = contents._rng_state
            del contents._rng_state
            return contents

        # QuestPlus.
        if sys.version_info.major == 3 and sys.version_info.minor >= 6:
            from psychopy.data.staircase import QuestPlusHandler
            from questplus import QuestPlus
            if isinstance(contents, QuestPlusHandler):
                # Restore the questplus.QuestPlus object.
                contents._qp = QuestPlus.from_json(contents._qp_json)
                del contents._qp_json
                return contents

        # If we haven't returned anything by now, the loaded object is neither
        # a TrialHandler2 nor a QuestPlus instance. Return it unchanged.
        return contents
    else:
        msg = "Don't know how to handle this file type, aborting."
        raise ValueError(msg)


def mergeFolder(src, dst, pattern=None):
    """Merge a folder into another.

    Existing files in `dst` folder with the same name will be
    overwritten. Non-existent files/folders will be created.
    """
    # dstdir must exist first
    srcnames = os.listdir(src)
    for name in srcnames:
        srcfname = os.path.join(src, name)
        dstfname = os.path.join(dst, name)
        if os.path.isdir(srcfname):
            if not os.path.isdir(dstfname):
                os.makedirs(dstfname)
            mergeFolder(srcfname, dstfname)
        else:
            try:
                # copy without metadata:
                shutil.copyfile(srcfname, dstfname)
            except IOError as why:
                print(why)


def openOutputFile(fileName=None, append=False, fileCollisionMethod='rename',
                   encoding='utf-8-sig'):
    """Open an output file (or standard output) for writing.

    :Parameters:

    fileName : None, 'stdout', or str
        The desired output file name. If `None` or `stdout`, return
        `sys.stdout`. Any other string will be considered a filename.
    append : bool, optional
        If ``True``, append data to an existing file; otherwise, overwrite
        it with new data.
        Defaults to ``True``, i.e. appending.
    fileCollisionMethod : string, optional
        How to handle filename collisions. Valid values are `'rename'`,
        `'overwrite'`, and `'fail'`.
        This parameter is ignored if ``append``  is set to ``True``.
        Defaults to `rename`.
    encoding : string, optional
        The encoding to use when writing the file. This parameter will be
        ignored if `append` is `False` and `fileName` ends with `.psydat`
        or `.npy` (i.e. if a binary file is to be written).
        Defaults to ``'utf-8'``.

    :Returns:

    f : file
        A writable file handle.

    """
    fileName = pathToString(fileName)
    if (fileName is None) or (fileName == 'stdout'):
        return sys.stdout

    if append:
        mode = 'a'
    else:
        if fileName.endswith(('.psydat', '.npy')):
            mode = 'wb'
        else:
            mode = 'w'

        # Rename the output file if a file of that name already exists
        # and it should not be appended.
        if os.path.exists(fileName) and not append:
            fileName = handleFileCollision(
                fileName,
                fileCollisionMethod=fileCollisionMethod)

    # Do not use encoding when writing a binary file.
    if 'b' in mode:
        encoding = None

    if os.path.exists(fileName) and mode in ['w', 'wb']:
        logging.warning('Data file %s will be overwritten!' % fileName)

    # The file will always be opened in binary writing mode,
    # see https://docs.python.org/2/library/codecs.html#codecs.open
    f = codecs.open(fileName, mode=mode, encoding=encoding)
    return f


def genDelimiter(fileName):
    """Return a delimiter based on a filename.

    :Parameters:

    fileName : string
        The output file name.

    :Returns:

    delim : string
        A delimiter picked based on the supplied filename. This will be
        ``,`` if the filename extension is ``.csv``, and a tabulator
        character otherwise.

    """
    fileName = pathToString(fileName)
    if fileName.endswith(('.csv', '.CSV')):
        delim = ','
    else:
        delim = '\t'

    return delim


def genFilenameFromDelimiter(filename, delim):
    # If no known filename extension was specified, derive a one from the
    # delimiter.
    filename = pathToString(filename)
    if not filename.endswith(('.dlm', '.DLM', '.tsv', '.TSV', '.txt',
                              '.TXT', '.csv', '.CSV', '.psydat', '.npy',
                              '.json')):
        if delim == ',':
            filename += '.csv'
        elif delim == '\t':
            filename += '.tsv'
        else:
            filename += '.txt'

    return filename


class DictStorage(dict):
    """Helper class based on dictionary with storage to json
    """

    def __init__(self, filename, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.filename = filename
        self.load()
        self._deleted = False
        atexit.register(self.__del__)

    def load(self, filename=None):
        """Load all tokens from a given filename
        (defaults to ~/.PsychoPy3/pavlovia.json)
        """
        if filename is None:
            filename = self.filename
        if os.path.isfile(filename):
            with open(filename, 'r') as f:
                try:
                    self.update(json.load(f))
                except ValueError:
                    logging.error("Tried to load %s but it wasn't valid "
                                  "JSON format"
                                  %filename)

    def save(self, filename=None):
        """Save all tokens from a given filename
        (defaults to the filename given to the class but can be overridden)
        """
        if filename is None:
            filename = self.filename
        # make sure the folder exists
        folder = os.path.split(filename)[0]
        if not os.path.isdir(folder):
            os.makedirs(folder)
        # save the file as json
        with open(filename, 'wb') as f:
            json_str = json.dumps(self, indent=2, sort_keys=True)
            f.write(bytes(json_str, 'UTF-8'))

    def __del__(self):
        if not self._deleted:
            self.save()
        self._deleted = True


class KnownProjects(DictStorage):
    def save(self, filename=None):
        """Purge unnecessary projects (without a local root) and save"""
        toPurge = []
        for projname in self:
            proj = self[projname]
            if not proj['localRoot']:
                toPurge.append(projname)
        for projname in toPurge:
            del self[projname]
        DictStorage.save(self, filename)


def pathToString(filepath):
    """
    Coerces pathlib Path objects to a string (only python version 3.6+)
    any other objects passed to this function will be returned as is.
    This WILL NOT work with on Python 3.4, 3.5 since the __fspath__ under
    method did not exist in those versions, however psychopy does not support
    these versions of python anyways.

    :Parameters:

    filepath : str or pathlib.Path
        file system path that needs to be coerced into a string to
        use by Psychopy's internals

    :Returns:

    filepath : str or same as input object
        file system path coerced into a string type
    """
    if hasattr(filepath, "__fspath__"):
        return filepath.__fspath__()
    return filepath



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""A Backend class defines the core low-level functions required by a Window
class, such as the ability to create an OpenGL context and flip the window.

Users simply call visual.Window(..., winType='pyglet') and the winType is then
used by backends.getBackend(winType) which will locate the appropriate class
and initialize an instance using the attributes of the Window.
"""

import weakref
from abc import ABC, abstractmethod

import numpy as np
from psychopy import logging
from psychopy.tools.attributetools import attributeSetter


class BaseBackend(ABC):
    """The backend abstract base class that defines all the core low-level
    functions required by a :class:`~psychopy.visual.Window` class.

    Such methods as the ability to create an OpenGL context, process events,
    and flip the window are prototyped here. Sub-classes of this function must
    implement the abstract methods shown here to be complete.

    Users simply call visual.Window(..., winType='pyglet') and the `winType` is
    then used by `backends.getBackend(winType)` which will locate the
    appropriate class and initialize an instance using the attributes of the
    Window.

    """
    # define GL here as a class attribute that includes all the opengl funcs
    # e.g. GL = pyglet.gl

    # define the name of the backend, used to register the name to use when
    # specifying `winType`
    # e.g. winTypeName = 'custom'

    def __init__(self, win):
        """Set up the backend window according the params of the PsychoPy win

        Before PsychoPy 1.90.0 this code was executed in Window._setupPyglet()

        :param: win is a PsychoPy Window (usually not fully created yet)
        """
        self.win = win  # this will use the @property to make/use a weakref

        # callback functions
        self._onMoveCallback = None
        self._onResizeCallback = None

        super().__init__()

    @abstractmethod
    def swapBuffers(self):
        """Set the gamma table for the graphics card

        """
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    @abstractmethod
    def setCurrent(self):
        """Sets this window to be the current rendering target (for backends
        where 2 windows are permitted, e.g. not pygame)
        """
        pass

    @attributeSetter
    def gamma(self, gamma):
        """Set the gamma table for the graphics card

        :param gamma: a single value or a triplet for separate RGB gamma values
        """
        self.__dict__['gamma'] = gamma
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    @attributeSetter
    def gammaRamp(self, gammaRamp):
        """Gets the gamma ramp or sets it to a new value (an Nx3 or Nx1 array)
        """
        self.__dict__['gammaRamp'] = gammaRamp
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    @property
    def shadersSupported(self):
        """This is a read-only property indicating whether or not this backend
        supports OpenGL shaders"""
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    # Optional, depending on backend needs

    def dispatchEvents(self):
        """This method is not needed for all backends but for engines with an
        event loop it may be needed to pump for new events (e.g. pyglet)
        """
        logging.warning("dispatchEvents() method in {} was called "
                        "but is not implemented. Is it needed?"
                        .format(self.win.winType)
                        )

    @property
    def onMoveCallback(self):
        """Callback function for window move events (`callable` or `None`).

        Callback function must have the following signature::

            callback(Any: winHandle, int: newPosX, int: newPosY) -> None

        """
        return self._onMoveCallback

    @onMoveCallback.setter
    def onMoveCallback(self, value):
        if not (callable(value) or value is None):
            raise TypeError(
                'Value for `onMoveCallback` must be callable or `None`.')

        self._onMoveCallback = value

    @property
    def onResizeCallback(self):
        """Callback function for window resize events (`callable` or `None`).

        Callback function must have the following signature::

            callback(Any: winHandle, int: newSizeW, int: newSizeH) -> None

        """
        return self._onResizeCallback

    @onResizeCallback.setter
    def onResizeCallback(self, value):
        if not (callable(value) or value is None):
            raise TypeError(
                'Value for `onResizeCallback` must be callable or `None`.')

        self._onResizeCallback = value

    def onResize(self, width, height):
        """A method that will be called if the window detects a resize event.

        This method is bound to the window backend resize event, data is
        formatted and forwarded to the user's callback function.

        """
        # When overriding this function, at the very minimum we must call the
        # user's function, passing the data they expect.
        if self._onResizeCallback is not None:
            self._onResizeCallback(self.win, width, height)

    def onMove(self, posX, posY):
        """A method called when the window is moved by the user.

        This method is bound to the window backend move event, data is
        formatted and forwarded to the user's callback function.

        """
        if hasattr(self.win, 'pos'):  # write the new position of the window
            self.win.pos = (posX, posY)

        if self._onMoveCallback is not None:
            self._onMoveCallback(self.win, posX, posY)

    # Helper methods that don't need converting

    @property
    def win(self):
        """The PsychoPy Window that this backend is supporting, which provides
        various important variables (like size, units, color etc).

        NB win is stored as a weakref to a psychopy.window and this property
        helpfully converts it back to a regular object so you don't need to
        think about it!
        """
        ref = self.__dict__['win']
        return ref()

    @win.setter
    def win(self, win):
        """The PsychoPy Window that this backend is supporting, which provides
        various important variables (like size, units, color etc).

        NB win is stored as a weakref to a psychopy.window and this property
        helpfully converts it back to a regular object so you don't need to
        think about it!
        """
        self.__dict__['win'] = weakref.ref(win)

    @property
    def autoLog(self):
        """If the window has logging turned on then backend should too"""
        return self.win.autoLog

    @property
    def name(self):
        """Name of the backend is only used for logging purposes"""
        return "{}_backend".format(self.win.name)

    # --------------------------------------------------------------------------
    # Window unit conversion
    #

    def _windowToBufferCoords(self, pos):
        """Convert window coordinates to OpenGL buffer coordinates.

        The standard convention for window coordinates is that the origin is at
        the top-left corner. The `y` coordinate increases in the downwards
        direction. OpenGL places the origin at bottom left corner, where `y`
        increases in the upwards direction.

        Parameters
        ----------
        pos : ArrayLike
            Position `(x, y)` in window coordinates.

        Returns
        -------
        ndarray
            Position `(x, y)` in buffer coordinates.

        """
        # This conversion is typical for many frameworks. If the framework uses
        # some other convention, that backend class should override this method
        # to ensure `_windowToPixCoords` returns the correct value.
        #
        invSf = 1.0 / self.win.getContentScaleFactor()
        return np.array(
            (pos[0] * invSf, (self.win.size[1] - pos[1]) * invSf),
            dtype=np.float32)

    def _bufferToWindowCoords(self, pos):
        """OpenGL buffer coordinates to window coordinates.

        This is the inverse of `_windowToBufferCoords`.

        Parameters
        ----------
        pos : ArrayLike
            Position `(x, y)` in window coordinates.

        Returns
        -------
        ndarray
            Position `(x, y)` in buffer coordinates.

        """
        # This conversion is typical for many frameworks. If the framework uses
        # some other convention, that backend class should override this method
        # to ensure `_windowToPixCoords` returns the correct value.
        #
        sf = self.win.getContentScaleFactor()
        return np.array(
            (pos[0] * sf, -pos[1] * sf + self.win.size[1]),
            dtype=np.float32)

    def _windowCoordsToPix(self, pos):
        """Convert window coordinates to the PsychoPy 'pix' coordinate system.
        This puts the origin at the center of the window.

        Parameters
        ----------
        pos : ArrayLike
            Position `(x, y)` in window coordinates.

        Returns
        -------
        ndarray
            Position `(x, y)` in PsychoPy pixel coordinates.

        """
        return np.asarray(self._windowToBufferCoords(pos) - self.win.size / 2.0,
                          dtype=np.float32)

    def _pixToWindowCoords(self, pos):
        """Convert PsychoPy 'pix' to the window coordinate system. This is the
        inverse of `_windowToPixCoords`.

        Parameters
        ----------
        pos : ArrayLike
            Position `(x, y)` in PsychoPy pixel coordinates.

        Returns
        -------
        ndarray
            Position `(x, y)` in window coordinates.

        """
        return self._bufferToWindowCoords(
            np.asarray(pos, dtype=np.float32) + self.win.size / 2.0)

    # --------------------------------------------------------------------------
    # Mouse related methods (e.g., event handlers)
    #
    # These methods are used to handle mouse events. Each function is bound to
    # the appropriate callback which registers the mouse event with the global
    # mouse event handler (psychopy.hardware.mouse.Mouse). Each callback has an
    # `*args` parameter which allows the backend to pass whatever parameters.
    #

    @abstractmethod
    def onMouseButton(self, *args, **kwargs):
        """Event handler for any mouse button event (pressed and released).

        This is used by backends which combine both button state changes into
        a single event. Usually this would pass events to the appropriate
        `onMouseButtonPress` and `onMouseButtonRelease` methods.
        """
        raise NotImplementedError(
            "`onMouseButton` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseButtonPress(self, *args, **kwargs):
        """Event handler for mouse press events. This handler can also be used
        for release events if the backend passes all button events to the same
        callback.
        """
        raise NotImplementedError(
            "`onMouseButtonPress` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseButtonRelease(self, *args, **kwargs):
        """Event handler for mouse release events."""
        raise NotImplementedError(
            "`onMouseButtonRelease` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseScroll(self, *args, **kwargs):
        """Event handler for mouse scroll events. Called when the mouse scroll
        wheel is moved."""
        raise NotImplementedError(
            "`onMouseScroll` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseMove(self, *args, **kwargs):
        """Event handler for mouse move events."""
        raise NotImplementedError(
            "`onMouseMove` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseEnter(self, *args, **kwargs):
        """Event called when the mouse enters the window. Some backends might
        combine enter and leave events to the same callback, this will handle
        both if so.
        """
        raise NotImplementedError(
            "`onMouseEnter` is not yet implemented for this backend.")

    @abstractmethod
    def onMouseLeave(self, *args, **kwargs):
        """Event called when a mouse leaves the window."""
        raise NotImplementedError(
            "`onMouseLeave` is not yet implemented for this backend.")

    @abstractmethod
    def getMousePos(self):
        """Get the position of the mouse on the current window.

        Returns
        -------
        ndarray
            Position `(x, y)` in window coordinates.

        """
        raise NotImplementedError(
            "`getMousePos` is not yet implemented for this backend.")

    @abstractmethod
    def setMousePos(self, pos):
        """Set/move the position of the mouse on the current window.

        Parameters
        ----------
        pos : ArrayLike
            Position `(x, y)` in window coordinates.

        """
        raise NotImplementedError(
            "`setMousePos` is not yet implemented for this backend.")

    def setMouseType(self, name='arrow'):
        """Change the appearance of the cursor for this window. Cursor types
        provide contextual hints about how to interact with on-screen objects.

        **Deprecated!** Use `setMouseCursor` instead.

        Parameters
        ----------
        name : str
            Type of standard cursor to use.

        """
        self.setMouseCursor(name)

    @abstractmethod
    def setMouseCursor(self, cursorType='default'):
        """Change the appearance of the cursor for this window. Cursor types
        provide contextual hints about how to interact with on-screen objects.

        The graphics used 'standard cursors' provided by the operating system.
        They may vary in appearance and hot spot location across platforms. The
        following names are valid on most platforms:

        * ``arrow`` or ``default`` : Default system pointer.
        * ``ibeam`` or ``text`` : Indicates text can be edited.
        * ``crosshair`` : Crosshair with hot-spot at center.
        * ``hand`` : A pointing hand.
        * ``hresize`` : Double arrows pointing horizontally.
        * ``vresize`` : Double arrows pointing vertically.
        * ``help`` : Arrow with a question mark beside it (Windows only).
        * ``no`` : 'No entry' sign or circle with diagonal bar.
        * ``size`` : Vertical and horizontal sizing.
        * ``downleft`` or ``upright`` : Double arrows pointing diagonally with
          positive slope (Windows only).
        * ``downright`` or ``upleft`` : Double arrows pointing diagonally with
          negative slope (Windows only).
        * ``lresize`` : Arrow pointing left (Mac OS X only).
        * ``rresize`` : Arrow pointing right (Mac OS X only).
        * ``uresize`` : Arrow pointing up (Mac OS X only).
        * ``dresize`` : Arrow pointing down (Mac OS X only).
        * ``wait`` : Hourglass (Windows) or watch (Mac OS X) to indicate the
           system is busy.
        * ``waitarrow`` : Hourglass beside a default pointer (Windows only).

        In cases where a cursor is not supported, the default for the system
        will be used.

        Parameters
        ----------
        cursorType : str
            Type of standard cursor to use. If not specified, `'default'` is
            used.

        Notes
        -----
        * On some platforms the 'crosshair' cursor may not be visible on uniform
          grey backgrounds.

        """
        raise NotImplementedError(
            "`setMouseCursor` is not yet implemented for this backend.")

    @abstractmethod
    def setMouseVisibility(self, visible):
        """Set mouse visibility.

        Parameters
        ----------
        visible : bool
            Mouse visibility mode.

        """
        raise NotImplementedError(
            "`setMouseVisibility` is not yet implemented for this backend.")

    @abstractmethod
    def setMouseExclusive(self, exclusive):
        """Set mouse exclusivity.

        Parameters
        ----------
        exclusive : bool
            Mouse exclusivity mode.

        """
        raise NotImplementedError(
            "`setMouseExclusive` is not yet implemented for this backend.")


if __name__ == "__main__":
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from os import path
from pathlib import Path
from psychopy.experiment.components import BaseComponent, Param, _translate
from psychopy.localization import _localized as __localized
import serial.tools.list_ports
_localized = __localized.copy()

_localized.update({'pulseDuration': _translate('Pulse duration (s)'),
                    'numberOfPulses': _translate('Number of pulses'),
                    'numberOfSequences': _translate('Number of sequences'),
                    'delayBetweenSeq': _translate('Delay between sequences (s)'),
                    'delayBetweenPulses': _translate('Delay between pulses (s)'),
                    'saveStats': _translate('Save actions of pump and licks to txt file'),
                    'stopVal': _translate('Duration (s)'),
                    'com_port': _translate('COM port'),})


class PeristalticPumpComponent(BaseComponent):
    """Delivers a water reward to the animal and monitor licks"""
    targets = ['PsychoPy']
    categories = ['I/O']
    iconFile = Path(__file__).parent / 'reward.png'
    tooltip = _translate(
        'LabeoTech Pump: Delivers a water reward to the animal and monitor '
        'water consumption (licks)')

    def __init__(self, exp, parentName, name='reward',
                 pulseDuration=0, numberOfPulses = 0, delayBetweenSeq = 0,
                 numberOfSequences = 0, delayBetweenPulses = 0,
                 startType='time (s)', startVal='0.0', stopVal='1.0',
                 stopType='duration (s)', saveStats = False,
                 com_port="Select pump com port"):

        super(PeristalticPumpComponent, self).__init__(
            exp, parentName, name, startType=startType, startVal=startVal,
            stopType=stopType, stopVal=stopVal
            )

        self.type = 'PeristalticPump'
        self.url = 'file:///C:/Users/delam/Desktop/BehavioralTask/pompe_sequence.html'
        # TODO : create a html help page with pump sequence on the options panel
        #self.url = Path(__file__).parent / 'pompe_sequence.html'

        # Order in which the user-settable parameters will be displayed
        # in the component's properties window.
        #self.order += ['PulseDuration',  # Basic tab]

        self.params['pulseDuration'] = Param(
            pulseDuration, categ='Basic',
            valType='num', inputType="single",
            hint=_translate('The duration of the pulse sent to the peristaltic pump'),
            label=_localized['pulseDuration'])

        self.params['numberOfPulses'] = Param(
            numberOfPulses, categ='Basic',
            valType='num', inputType="single",
            hint=_translate('Number of pulses in a burst sequence'),
            label=_localized['numberOfPulses'])

        self.params['delayBetweenSeq'] = Param(
            delayBetweenSeq, categ='Basic',
            valType='num', inputType="single",
            hint=_translate('Delay between sequences'),
            label=_localized['delayBetweenSeq'])

        self.params['numberOfSequences'] = Param(
            numberOfSequences, categ='Basic',
            valType='num', inputType="single",
            hint=_translate('Number of sequence in a single reward event'),
            label=_localized['numberOfSequences'])

        self.params['delayBetweenPulses'] = Param(
            delayBetweenPulses, categ='Basic',
            valType='num', inputType="single",
            hint=_translate('Delay between pulses in a burst sequence'),
            label=_localized['delayBetweenPulses'])

        self.params['saveStats'] = Param(
            saveStats, categ='Basic',
            valType='bool', inputType="bool",
            hint=_translate('Save log to txt file'),
            label=_localized['saveStats'])

        self.params['stopVal'] = Param(
            ((((pulseDuration * numberOfPulses + delayBetweenPulses) *
               (numberOfPulses - 1)) * numberOfSequences) +
             (delayBetweenSeq * (numberOfSequences - 1))),
            categ='Basic', valType='num', inputType="single",
            hint=_translate('The duration of the pulse sent to the peristaltic pump'),
            label=_localized['stopVal'])
            
        self.params['com_port'] = Param(
            com_port, valType='str', inputType="choice", categ='Basic',
            allowedVals=[p.device for p in serial.tools.list_ports.comports()],
            updates='constant',
            hint=_translate("COM port"),
            label=_localized['com_port'])


    def writeInitCode(self, buff):
            
        """Write variable initialisation code."""
        code =  ("%(name)s = event.Mouse(win=win)\n")
        code += ("pulse_dur = %(pulseDuration)s\n")
        code+= ("number_pulses = %(numberOfPulses)s\n")
        code+= ("delay_sequences = %(delayBetweenSeq)s\n")
        code+= ("number_sequences = %(numberOfSequences)s\n")
        code+= ("delay_pulses = %(delayBetweenPulses)s\n")
        code+= ("saveStats = %(saveStats)s\n")
        code+= ("import psychopy\n")
        code+= ("import serial\n")
        code+= ("import time\n")
        code+=("from datetime import datetime\n")
        code+= ("pump = serial.Serial(port=%(com_port)s, baudrate=115200, timeout=.1)\n")
        code+= ("pulse_started = False\n")
        code+= ("reward_start = %(startVal)s\n")
        code+=("n=0\n")
        code+=("ns=0\n")
        code+=("tp0=0\n")
        code+=("td0=0\n")
        code+=("ts0=0\n")
        code+=("firstFrame = False\n")
        code+=("pause = False\n")
        code+=("""expInfo = {'participant': '', 'session': '001'}\n""")
        code+=("""expInfo['date'] = data.getDateStr()\n""")
        buff.writeIndented(code % self.params)
        code=("""filename_pump = _thisDir + os.sep + u'data/%s_pump_%s' % (expInfo['participant'], expInfo['date'])\n""")
        buff.writeIndented(code)
        code=("""text_file_pump = open(filename_pump + '.txt', 'w')\n""")
        buff.writeIndented(code)


    def writeRoutineStartCode(self, buff):
        """Write the code that will be called at the start of the routine."""
        pass

    def writeFrameCode(self, buff):
        
        buff.writeIndented("if pump.inWaiting():\n")
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("lick_state = pump.readline()\n")
        buff.writeIndented("""if lick_state == b'l\\r\\n':\n""")
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("try:\n")
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("if behavioral_sys == True:\n")
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("behavioral_sys = True\n")
        buff.setIndentLevel(-2, relative=True)
        buff.writeIndented("except:\n")
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("""print(str(round(t,2)) + ': lick')\n""")
        buff.writeIndented("if saveStats == True:\n")
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("""text_file_pump.write(str(round(t,2)) + 'lick')\n""")
        #code+=("""    lick_state = b''\n""")
        buff.setIndentLevel(-4, relative=True)
        
        buff.writeIndented("if t>= %(startVal)s and %(name)s.status != FINISHED:\n" % self.params)
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("%(name)s.status = STARTED\n" % self.params)
        buff.setIndentLevel(-1, relative=True)
        buff.writeIndented("if firstFrame == False and  %(name)s.status == STARTED:\n" % self.params)
        buff.setIndentLevel(1, relative=True)

        buff.writeIndented("firstFrame = True\n")
        buff.writeIndented("tp0 = time.time()\n")
        buff.writeIndented("td=0\n")
        buff.writeIndented("td0=time.time()\n")
        buff.writeIndented("ts0=time.time()\n")
        buff.writeIndented("ts=0\n")
        buff.writeIndented("""pump.write(bytes('o', 'utf-8'))\n""")
        buff.writeIndented("""print(str(round(t,2)) + ': pump ON')\n""")
        buff.writeIndented("pulse_started = True\n")

        buff.setIndentLevel(-1, relative=True)
        buff.writeIndented("tp = time.time() - tp0\n")
        buff.writeIndented("ts = time.time() - ts0\n")
        buff.writeIndented("if tp>= pulse_dur and pulse_started == True and pause == False and firstFrame == True and %(name)s.status == STARTED:\n" % self.params)
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("""pump.write(bytes('f', 'utf-8'))\n""")
        buff.writeIndented("""print(str(round(t,2)) + ': pump OFF')\n""")
        buff.writeIndented("pulse_started = False\n")
        buff.writeIndented("n+=1\n")
        buff.writeIndented("td0 = time.time()\n")

        buff.setIndentLevel(-1, relative=True)
        buff.writeIndented("td = time.time() - td0\n")
        buff.writeIndented("if td >= delay_pulses and pulse_started == False and pause == False and firstFrame == True:\n")
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("pulse_started = True\n")
        buff.writeIndented("if not n >= number_pulses :\n")
        buff.setIndentLevel(1, relative=True)

        buff.writeIndented("""pump.write(bytes('o', 'utf-8'))\n""")
        buff.writeIndented("""print(str(round(t,2)) + ': pump ON')\n""")
        buff.writeIndented("pulse_started = True\n")
        buff.writeIndented("tp0 = time.time()\n")
        buff.setIndentLevel(-2, relative=True)
        buff.writeIndented("if n == number_pulses and %(name)s.status == STARTED and pause == False:\n" % self.params)
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("""pump.write(bytes('f', 'utf-8'))\n""")
        buff.writeIndented("ts0 = time.time()\n")
        buff.writeIndented("pause=True\n")
        buff.writeIndented("ns+=1\n")
        buff.writeIndented("n=0\n")
        buff.setIndentLevel(-1, relative=True)
        buff.writeIndented("if ns == number_sequences and %(name)s.status == STARTED:\n" % self.params)
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("pump.write(bytes('f', 'utf-8'))\n")
        buff.writeIndented("%(name)s.status = FINISHED\n" % self.params)
        buff.setIndentLevel(-1, relative=True)
        buff.writeIndented("ts = time.time()-ts0\n")
        buff.writeIndented("if ts >= delay_sequences and pause == True and firstFrame == True:\n")
        buff.setIndentLevel(1, relative=True)
        buff.writeIndented("pause=False\n")
        buff.writeIndented("pulse_started = True\n")
        buff.writeIndented("""pump.write(bytes('o', 'utf-8'))\n""")
        buff.writeIndented("""print(str(round(t,2)) + ': pump ON')\n""")
        buff.writeIndented("tp0 = time.time()\n")
    
        buff.setIndentLevel(-1, relative=True)

    def writeRoutineEndCode(self, buff):
        buff.writeIndented("%(name)s.status = 0\n" % self.params)
        buff.writeIndented("ns = 0\n")
        buff.writeIndented("n = 0\n")
        buff.writeIndented("firstFrame = False\n")
        buff.writeIndented("pause = False\n")


    def writeExperimentEndCode(self, buff):
        pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Demo using the new (beta) VlcMovieStim to play a video file. Path of video
needs to updated to point to a video you have. The demo closes automatically at
the end of the stream.

This requires:

1. VLC. Just install the standard VLC of the same bitness as Python for your OS.

    https://www.videolan.org/vlc/index.html

2. pip install python-vlc

"""

import os
from psychopy import visual, core, event

# get the video from the demo resources directory
videopath = r'jwpIntro.mp4'
videopath = os.path.join(os.getcwd(), videopath)
if not os.path.exists(videopath):
    raise RuntimeError("Video File could not be found:" + videopath)

# open a window to display the video
win = visual.Window([800, 800], fullscr=False)

# create text stim for instructions
keystext = "PRESS 'q' or 'escape' to Quit.\n"
keystext += "'s': Stop and restart\n"
keystext += "'p': Pause/unpause\n"
keystext += "'>': Seek forward 1 second\n"
keystext += "'<': Seek backward 1 second\n"
keystext += "'-': Decrease volume\n"
keystext += "'+': Increase volume"
text = visual.TextStim(win, keystext, pos=(0, -250), units='pix')

# Create your movie stim
mov = visual.VlcMovieStim(win, videopath,
    size=600,  # set as `None` to use the native video size
    pos=[0, 0],  # pos specifies the /center/ of the movie stim location
    flipVert=False,  # flip the video picture vertically
    flipHoriz=False,  # flip the video picture horizontally
    loop=False,  # replay the video when it reaches the end
    autoStart=True)  # start the video automatically when first drawn

# main loop, will exit automatically when the movie is finished
while not mov.isFinished:
    # Check for action keys.....
    for key in event.getKeys():
        if key in ['escape', 'q']:
            mov.stop()
            core.quit()
        elif key in ['s', ]:
            mov.replay()
        elif key in ['p', ]:
            # To pause the movie while it is playing ...
            if mov.isNotStarted or mov.isPaused:
                mov.play()
            elif mov.isPlaying:
                mov.pause()
        elif key == 'period':
            # To skip ahead 1 second in movie ...
            mov.fastForward(1.0)
        elif key == 'comma':
            # To skip back 1 second in movie ...
            mov.rewind(1.0)
        elif key == 'minus':
            # To decrease movie sound a bit (5%) ...
            mov.decreaseVolume(5)
        elif key == 'equal':
            # To increase movie sound a bit (5%) ...
            mov.increaseVolume(5)

    # draw elements
    mov.draw()  # movie frame
    text.draw()  # instruction text overlay
    win.flip()  # flip buffers to draw the content to the window

win.close()
core.quit()

# The contents of this file are in the public domain.



"""
The Alert module
"""
from ._alerts import alert, catalog

import yaml
from pathlib import Path
import re


def validateCatalogue(dev=False):
    """
    Check that every value in the catalogue corresponds to a yaml file.

    dev : If True, then missing entries will be created from template. Useful when writing alerts.
    """
    # Define root folder
    root = Path(__file__).parent / "alertsCatalogue"
    # Make template object
    with open(root / "alertTemplate.yaml") as f:
        template = f.read()
    # Create blank array to store missing alert keys in
    missing = []

    def validate(spec):
        # Start off valid
        valid = True
        # Get category
        if "cat" in spec:
            cat = spec['cat']
        else:
            cat = "Unknown"
        for key, val in spec.items():
            if key == "cat" or key % 1000 == 0 or key == 9999:
                # Skip category tags
                continue
            if isinstance(val, str):
                # Check whether file exists
                file = root / f"{key}.yaml"
                valid = valid and file.is_file()
                # If in dev mode, make a new yaml file from template
                if dev and not file.is_file():
                    newAlert = template
                    # Replace whatever is possible at this stage
                    newAlert = re.sub(r"(?<=code\: )\d\d\d\d", str(key), newAlert)
                    newAlert = re.sub(r"(?<=cat\: ).*", str(cat), newAlert)
                    newAlert = re.sub(r"(?<=label\: ).*", str(val), newAlert)
                    # Save new file
                    with open(file, "w") as f:
                        f.write(newAlert)
                        f.close()
            # Store missing key
            if not valid:
                missing.append(key)
            if isinstance(val, dict):
                # Recursively search through dicts
                valid = valid and validate(val)
        return valid

    # Load catalog
    with open(root / "alertCategories.yaml") as f:
        spec = yaml.load(f, Loader=yaml.FullLoader)
    # Validate
    valid = validate(spec)
    return valid, missing



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from pathlib import Path

from psychopy.experiment.components import Param, _translate
from psychopy.experiment.components.keyboard import KeyboardComponent
from psychopy.experiment import CodeGenerationException, valid_var_re
from psychopy.localization import _localized as __localized
_localized = __localized.copy()
__author__ = 'Jon Peirce'

# only use _localized values for label values, nothing functional:
_localized.update({'deviceNumber': _translate('Device number'),
                   'useBoxTimer': _translate("Use box timer")})


class cedrusButtonBoxComponent(KeyboardComponent):
    """An event class for checking an Cedrus RBxxx button boxes
    using XID library

    This is based on keyboard component, several important differences:
    - no special response class analogous to event.BuilderKeyResponse()
    - enabled responses (active keys) are handled by the hardware device

    More than one component in a routine will produce conflicts between
    components over which active keys (for responses and lights).
    """
    categories = ['Responses']  # which section(s) in the components panel
    targets = ['PsychoPy']
    iconFile = Path(__file__).parent / 'cedrusBox.png'
    tooltip = _translate('Cedrus Button Box: Cedrus response boxes, using the '
                         'pyxid2 library provided by Cedrus')

    def __init__(self, exp, parentName, name='buttonBox',
                 store='first key',
                 useTimer=True, deviceNumber=0, allowedKeys="",
                 getReleaseTime=False,  # not yet supported
                 forceEndRoutine=True, storeCorrect=False, correctAns="",
                 discardPrev=True,
                 startType='time (s)', startVal=0.0,
                 stopType='duration (s)', stopVal=1.0,
                 startEstim='', durationEstim='',):
        super(cedrusButtonBoxComponent, self).__init__(
            exp, parentName, name=name,
            allowedKeys=allowedKeys, store=store, discardPrev=discardPrev,
            forceEndRoutine=forceEndRoutine, storeCorrect=storeCorrect,
            correctAns=correctAns, startType=startType, startVal=startVal,
            stopType=stopType, stopVal=stopVal,
            startEstim=startEstim, durationEstim=durationEstim)

        self.type = 'cedrusButtonBox'
        self.url = "https://www.psychopy.org/builder/components/cedrusButtonBox.html"
        self.order += ['forceEndRoutine',  # Basic tab
                       'allowedKeys', 'store', 'storeCorrect', 'correctAns'  # Data tab
                       ]

        self.exp.requirePsychopyLibs(['hardware'])

        self.params['correctAns'].hint = _translate(
            "What is the 'correct' response? NB, buttons are labelled 0 to "
            "6 on a 7-button box. Enter 'None' (no quotes) if withholding "
            "a response is correct. Might be helpful to add a correctAns "
            "column and use $correctAns to compare to the key press.")

        self.params['correctAns'].valType = 'code'

        self.params['allowedKeys'].hint = _translate(
            'Keys to be read (blank for any) or key numbers separated by '
            'commas')

        msg = _translate('Device number, if you have multiple devices which'
                         ' one do you want (0, 1, 2...)')
        self.params['deviceNumber'] = Param(
            deviceNumber, valType='int', inputType="spin", allowedTypes=[], categ='Hardware',
            updates='constant', allowedUpdates=[],
            hint=msg,
            label=_localized['deviceNumber'])

        # self.params['getReleaseTime'] = Param(getReleaseTime,
        #    valType='bool', allowedVals=[True, False],
        #    updates='constant', allowedUpdates=[],
        #    hint="Wait for the key to be released and store the time
        #       that it was held down",
        #    label="Get release time")

        msg = _translate('According to Cedrus the response box timer has '
                         'a drift - use with caution!')
        self.params['useBoxTimer'] = Param(
            getReleaseTime, valType='bool', inputType="bool", allowedVals=[True, False], categ='Hardware',
            updates='constant', allowedUpdates=[],
            hint=msg,
            label=_localized['useBoxTimer'])

    def writeRunOnceInitCode(self, buff):
        code = ("try:  # to use the Cedrus response box\n"
                "   import pyxid2 as pyxid\n"
                "except ImportError:\n"
                "   import pyxid\n"
                "cedrusBox_%(deviceNumber)s = None\n"
                "for n in range(10):  # doesn't always work first time!\n"
                "    try:\n"
                "        devices = pyxid.get_xid_devices()\n"
                "        core.wait(0.1)\n"
                "        cedrusBox_%(deviceNumber)s = devices[%(deviceNumber)s]\n"
                "        cedrusBox_%(deviceNumber)s.clock = core.Clock()\n"
                "        break  # found the device so can break the loop\n"
                "    except Exception:\n"
                "        pass\n"
                "if not cedrusBox_%(deviceNumber)s:\n"
                "    logging.error('could not find a Cedrus device.')\n"
                "    core.quit()\n")
        buff.writeOnceIndentedLines(code % self.params)

    def writeInitCode(self, buff):
        code = ("%(name)s = cedrusBox_%(deviceNumber)s\n")
        buff.writeIndentedLines(code % self.params)

    def writeRoutineStartCode(self, buff):
        if (self.params['store'].val != 'nothing' or
                self.params['storeCorrect'].val):
            code = ("%(name)s.keys = []  # to store response values\n"
                    "%(name)s.rt = []\n"
                    "%(name)s.status = None\n")
            buff.writeIndentedLines(code % self.params)

    def writeFrameCode(self, buff):
        """Write the code that will be called every frame.
        """
        # some shortcuts
        store = self.params['store'].val
        storeCorr = self.params['storeCorrect'].val
        forceEnd = self.params['forceEndRoutine'].val
        useBoxTimer = self.params['useBoxTimer'].val

        # check whether we need to test for allowed keys (or just take all)
        allowedKeys = self.params['allowedKeys'].val.strip()
        allowedKeysIsVar = valid_var_re.match(
            str(allowedKeys)) and not allowedKeys == 'None'
        if allowedKeysIsVar:
            # only insert this code if we think allowed keys is a variable.
            # check at run-time that the var is suitable to eval
            stringType = 'str'
            code = ("# AllowedKeys looks like a variable named `{0}`\n"
                    "if not '{0}' in locals():\n"
                    "    logging.error('AllowedKeys variable `{0}` "
                    "is not defined.')\n"
                    "    core.quit()\n" +
                    "if not type({0}) in [list, tuple, np.ndarray]:\n"
                    "    if not isinstance({0}, str):\n"
                    "        logging.error('AllowedKeys variable `{0}`"
                    " is not string- or list-like.')\n"
                    "        core.quit()\n" +
                    "    elif not ',' in {0}: {0} = ({0},)\n"
                    "    else:  {0} = eval({0})\n").format(allowedKeys, stringType)
            buff.writeIndentedLines(code)

            keyCheckStr = "keyList=list(%s)" % allowedKeys  # eval() @ run time
            keyList = allowedKeys

        # now create the string that will loop-continue if
        if allowedKeys in [None, "none", "None", "", "[]", "()"]:
            keyCheckStr = ""
        elif not allowedKeysIsVar:
            try:
                keyList = eval(allowedKeys)
            except Exception:
                raise CodeGenerationException(
                    self.params["name"], "Allowed keys list is invalid.")
            # this means the user typed "left","right" not ["left","right"]
            if type(keyList) == tuple:
                keyList = list(keyList)
            elif isinstance(keyList, str):  # a single string/key
                keyList = [keyList]
            keyCheckStr = "%s" % (repr(keyList))

        # if just now starting on this frame:
        buff.writeIndented("# *%(name)s* updates\n" % self.params)
        # write start code
        # writes an if statement to determine whether to start
        self.writeStartTestCode(buff)
        code = ("%(name)s.status = STARTED\n"
                "%(name)s.clock.reset()  # now t=0\n")
        buff.writeIndentedLines(code % self.params)

        if self.params['discard previous'].val:
            code = ("# clear %(name)s responses (in a loop - the Cedrus "
                    "own function doesn't work well)\n"
                    "%(name)s.poll_for_response()\n"
                    "while len(%(name)s.response_queue):\n"
                    "    %(name)s.clear_response_queue()\n"
                    "    %(name)s.poll_for_response() #often there are "
                    "more resps waiting!\n")
            buff.writeIndentedLines(code % self.params)

        if useBoxTimer:
            code = "%(name)s.reset_rt_timer()\n"
            buff.writeIndented(code % self.params)

        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)
        # test for stop (only if there was some setting for duration or stop)
        if self.params['stopVal'].val not in ['', None, -1, 'None']:
            # writes an if statement to determine whether to draw etc
            self.writeStopTestCode(buff)
            buff.writeIndented("%(name)s.status = FINISHED\n" % self.params)
            buff.setIndentLevel(-2, True)

        buff.writeIndented("if %(name)s.status == STARTED:\n" % self.params)
        buff.setIndentLevel(1, relative=True)  # to get out of if statement
        dedentAtEnd = 1  # keep track of how far to dedent later

        code = ("theseKeys=[]\n"
                "theseRTs=[]\n"
                "# check for key presses\n"
                "%(name)s.poll_for_response()\n"
                "while len(%(name)s.response_queue):\n"
                "    evt = %(name)s.get_next_response()\n")

        buff.writeIndentedLines(code % self.params)

        if len(keyCheckStr):
            code = ("    if evt['key'] not in %s:\n" % keyList +
                    "        continue  # we don't care about this key\n")
            buff.writeIndentedLines(code)

        code = ("    if evt['pressed']:\n"
                "      theseKeys.append(evt['key'])\n")
        buff.writeIndentedLines(code)

        if useBoxTimer:
            code = "      theseRTs.append(evt['time']/1000.0)\n"
            buff.writeIndented(code)
        else:
            code = "      theseRTs.append(%(name)s.clock.getTime())\n"
            buff.writeIndented(code % self.params)

        code = ("    %(name)s.poll_for_response()\n"
                "%(name)s.clear_response_queue()  # don't process again\n")
        buff.writeIndentedLines(code % self.params)

        # how do we store it?
        if store != 'nothing' or forceEnd:
            # we are going to store something
            code = "if len(theseKeys) > 0:  # at least one key was pressed\n"
            buff.writeIndented(code)
            buff.setIndentLevel(1, True)
            dedentAtEnd += 1  # indent by 1

        if store == 'first key':  # then see if a key has already been pressed
            code = "if %(name)s.keys == []:  # then this is first keypress\n"
            buff.writeIndented(code % self.params)

            buff.setIndentLevel(1, True)
            dedentAtEnd += 1  # indent by 1

            code = ("%(name)s.keys = theseKeys[0]  # the first key pressed\n"
                    "%(name)s.rt = theseRTs[0]\n")
            buff.writeIndentedLines(code % self.params)
        elif store == 'last key':
            code = ("%(name)s.keys = theseKeys[-1]  # the last key pressed\n"
                    "%(name)s.rt = theseRTs[-1]\n")
            buff.writeIndentedLines(code % self.params)
        elif store == 'all keys':
            code = ("%(name)s.keys.extend(theseKeys)  # storing all keys\n"
                    "%(name)s.rt.extend(theseRTs)\n")
            buff.writeIndentedLines(code % self.params)
        else:
            print((store, type(store), str(store)))
        if storeCorr:
            code = ("# was this 'correct'?\n"
                    "if (%(name)s.keys == str(%(correctAns)s)) or "
                    "(%(name)s.keys == %(correctAns)s):\n"
                    "    %(name)s.corr = 1\n"
                    "else:\n"
                    "    %(name)s.corr = 0\n")
            buff.writeIndentedLines(code % self.params)

        if forceEnd is True:
            code = ("# a response ends the routine\n"
                    "continueRoutine = False\n")
            buff.writeIndentedLines(code)
        buff.setIndentLevel(-(dedentAtEnd), relative=True)

    # this was commented-out (removed Feb 2016, available in history):
    # def writeRoutineEndCode(self, buff):



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from os import path
from pathlib import Path

from psychopy.alerts import alerttools
from psychopy.experiment.components import BaseVisualComponent, Param, getInitVals, _translate
from psychopy.experiment.py2js_transpiler import translatePythonToJavaScript
from psychopy.localization import _localized as __localized
_localized = __localized.copy()

# only use _localized values for label values, nothing functional:
_localized.update({'callback': _translate("Callback Function"),
                   'forceEndRoutine': _translate('Force end of Routine'),
                   'text': _translate('Button text'),
                   'font': _translate('Font'),
                   'letterHeight': _translate('Letter height'),
                   'bold': _translate('Bold'),
                   'italic': _translate('Italic'),
                   'padding': _translate('Padding'),
                   'anchor': _translate('Anchor'),
                   'fillColor': _translate('Fill Colour'),
                   'borderColor': _translate('Border Colour'),
                   'borderWidth': _translate('Border Width'),
                   'oncePerClick': _translate('Run once per click'),
                   'save': _translate("Record clicks"),
                   'timeRelativeTo': _translate("Time relative to")
                   })


class ButtonComponent(BaseVisualComponent):
    """
    A component for presenting a clickable textbox with a programmable callback
    """
    categories = ['Responses']
    targets = ['PsychoPy', 'PsychoJS']
    iconFile = Path(__file__).parent / 'button.png'
    tooltip = _translate('Button: A clickable textbox')
    beta = True

    def __init__(self, exp, parentName, name="button",
                 startType='time (s)', startVal=0,
                 stopType='duration (s)', stopVal=1.0,
                 startEstim='', durationEstim='',
                 text=_translate("Click here"), font='Arvo',
                 pos=(0, 0), size="", padding="", anchor='center', units='from exp settings', ori=0,
                 color="white", fillColor="darkgrey", borderColor="None", borderWidth=0, colorSpace='rgb', opacity="",
                 letterHeight=0.05, bold=True, italic=False,
                 callback="", save='every click', timeRelativeTo='button onset', forceEndRoutine=True, oncePerClick=True):
        super(ButtonComponent, self).__init__(exp, parentName, name,
                                            units=units,
                                            color=color, fillColor=fillColor, borderColor=borderColor,
                                            colorSpace=colorSpace,
                                            pos=pos,
                                            ori=ori,
                                            size=size,
                                            startType=startType,
                                            startVal=startVal,
                                            stopType=stopType,
                                            stopVal=stopVal,
                                            startEstim=startEstim,
                                            durationEstim=durationEstim)
        self.type = 'Button'
        self.url = "https://www.psychopy.org/builder/components/button.html"
        self.order += [  # controls order of params within tabs
            "forceEndRoutine", "text", "callback", "oncePerClick", # Basic tab
            "borderWidth", "opacity",  # Appearance tab
            "font", "letterHeight", "lineSpacing", "bold", "italic",  # Formatting tab
        ]
        # params
        _allow3 = ['constant', 'set every repeat', 'set every frame']  # list
        self.params['color'].label = _translate("Text Color")

        self.params['forceEndRoutine'] = Param(
            forceEndRoutine, valType='bool', inputType="bool", categ='Basic',
            updates='constant', direct=False,
            hint=_translate("Should a response force the end of the Routine "
                            "(e.g end the trial)?"),
            label=_localized['forceEndRoutine'])

        # If force end routine, then once per click doesn't make sense
        self.depends += [
            {
                "dependsOn": "forceEndRoutine",
                "condition": "==True",
                "param": "oncePerClick",
                "true": "disable",  # what to do with param if condition is True
                "false": "enable",  # permitted: hide, show, enable, disable
            }
        ]

        self.params['oncePerClick'] = Param(
            oncePerClick, valType='bool', inputType="bool", allowedTypes=[], categ='Basic',
            updates='constant',
            hint=_translate("Should the callback run once per click (True), or each frame until click is released (False)"),
            label=_localized['oncePerClick']
        )
        self.params['callback'] = Param(
            callback, valType='extendedCode', inputType="multi", allowedTypes=[], categ='Basic',
            updates='constant',
            hint=_translate("Code to run when button is clicked"),
            label=_localized['callback'])
        self.params['text'] = Param(
            text, valType='str', inputType="single", allowedTypes=[], categ='Basic',
            updates='constant', allowedUpdates=_allow3[:],  # copy the list
            hint=_translate("The text to be displayed"),
            label=_localized['text'])
        self.params['font'] = Param(
            font, valType='str', inputType="single", allowedTypes=[], categ='Formatting',
            updates='constant', allowedUpdates=_allow3[:],  # copy the list
            hint=_translate("The font name (e.g. Comic Sans)"),
            label=_localized['font'])
        self.params['letterHeight'] = Param(
            letterHeight, valType='num', inputType="single", allowedTypes=[], categ='Formatting',
            updates='constant', allowedUpdates=_allow3[:],  # copy the list
            hint=_translate("Specifies the height of the letter (the width"
                            " is then determined by the font)"),
            label=_localized['letterHeight'])
        self.params['italic'] = Param(
            italic, valType='bool', inputType="bool", allowedTypes=[], categ='Formatting',
            updates='constant',
            hint=_translate("Should text be italic?"),
            label=_localized['italic'])
        self.params['bold'] = Param(
            bold, valType='bool', inputType="bool", allowedTypes=[], categ='Formatting',
            updates='constant',
            hint=_translate("Should text be bold?"),
            label=_localized['bold'])
        self.params['padding'] = Param(
            padding, valType='num', inputType="single", allowedTypes=[], categ='Layout',
            updates='constant', allowedUpdates=_allow3[:],
            hint=_translate("Defines the space between text and the textbox border"),
            label=_localized['padding'])
        self.params['anchor'] = Param(
            anchor, valType='str', inputType="choice", categ='Layout',
            allowedVals=['center',
                         'top-center',
                         'bottom-center',
                         'center-left',
                         'center-right',
                         'top-left',
                         'top-right',
                         'bottom-left',
                         'bottom-right',
                         ],
            updates='constant',
            hint=_translate("Should text anchor to the top, center or bottom of the box?"),
            label=_localized['anchor'])
        self.params['borderWidth'] = Param(
            borderWidth, valType='num', inputType="single", allowedTypes=[], categ='Appearance',
            updates='constant', allowedUpdates=_allow3[:],
            hint=_translate("Textbox border width"),
            label=_localized['borderWidth'])
        self.params['save'] = Param(
            save, valType='str', inputType="choice", categ='Data',
            allowedVals=['first click', 'last click', 'every click', 'none'],
            hint=_translate(
                "What clicks on this button should be saved to the data output?"),
            direct=False,
            label=_localized['save'])
        self.params['timeRelativeTo'] = Param(
            timeRelativeTo, valType='str', inputType="choice", categ='Data',
            allowedVals=['button onset', 'experiment', 'routine'],
            updates='constant',
            direct=False,
            hint=_translate(
                "What should the values of mouse.time should be "
                "relative to?"),
            label=_localized['timeRelativeTo'])


    def writeInitCode(self, buff):
        # do we need units code?
        if self.params['units'].val == 'from exp settings':
            unitsStr = ""
        else:
            unitsStr = "units=%(units)s," % self.params
        # do writing of init
        inits = getInitVals(self.params, 'PsychoPy')
        code = (
                "%(name)s = visual.ButtonStim(win, \n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                    "text=%(text)s, font=%(font)s,\n"
                    "pos=%(pos)s," + unitsStr + "\n"
                    "letterHeight=%(letterHeight)s,\n"
                    "size=%(size)s, borderWidth=%(borderWidth)s,\n"
                    "fillColor=%(fillColor)s, borderColor=%(borderColor)s,\n"
                    "color=%(color)s, colorSpace=%(colorSpace)s,\n"
                    "opacity=%(opacity)s,\n"
                    "bold=%(bold)s, italic=%(italic)s,\n"
                    "padding=%(padding)s,\n"
                    "anchor=%(anchor)s,\n"
                    "name='%(name)s'\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
                ")\n"
                "%(name)s.buttonClock = core.Clock()"
        )
        buff.writeIndentedLines(code % inits)

    def writeInitCodeJS(self, buff):
        inits = getInitVals(self.params, 'PsychoJS')

        code = (
            "%(name)s = new visual.ButtonStim({\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                "win: psychoJS.window,\n"
                "name: '%(name)s',\n"
                "text: %(text)s,\n"
                "fillColor: %(fillColor)s,\n"
                "borderColor: %(borderColor)s,\n"
                "color: %(color)s,\n"
                "colorSpace: %(colorSpace)s,\n"
                "pos: %(pos)s,\n"
                "letterHeight: %(letterHeight)s,\n"
                "size: %(size)s\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            "});\n"
            "%(name)s.clock = new util.Clock();\n\n"
        )
        buff.writeIndentedLines(code % inits)

    def writeFrameCode(self, buff):
        BaseVisualComponent.writeFrameCode(self, buff)
        # do writing of init
        inits = getInitVals(self.params, 'PsychoPy')
        # Get callback from params
        callback = inits['callback']
        if inits['callback']:
            callback = str(callback)
        else:
            callback = ""
        # String to get time
        if inits['timeRelativeTo'] == 'button onset':
            timing = "%(name)s.buttonClock.getTime()"
        elif inits['timeRelativeTo'] == 'experiment':
            timing = "globalClock.getTime()"
        elif inits['timeRelativeTo'] == 'routine':
            timing = "routineTimer.getTime()"
        else:
            timing = "globalClock.getTime()"
        # Assemble code
        code = (
            f"if %(name)s.status == STARTED:\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                f"# check whether %(name)s has been pressed\n"
                f"if %(name)s.isClicked:\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                    f"if not %(name)s.wasClicked:\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                        f"%(name)s.timesOn.append({timing}) # store time of first click\n"
                        f"%(name)s.timesOff.append({timing}) # store time clicked until\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
                    f"else:\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                        f"%(name)s.timesOff[-1] = {timing} # update time clicked until\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        if self.params['oncePerClick'].val or self.params['forceEndRoutine'].val:
            code = (
                f"if not %(name)s.wasClicked:\n"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(1, relative=True)
            if self.params['forceEndRoutine'].val:
                code = (
                        f"continueRoutine = False  # end routine when %(name)s is clicked\n"
                )
                buff.writeIndentedLines(code % inits)
            if self.params['oncePerClick'].val:
                buff.writeIndentedLines(callback % inits)
                buff.setIndentLevel(-1, relative=True)
        if not self.params['oncePerClick'].val:
            buff.writeIndentedLines(callback % inits)
        code = (
                    f"%(name)s.wasClicked = True  # if %(name)s is still clicked next frame, it is not a new click\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
                f"else:\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                    f"%(name)s.wasClicked = False  # if %(name)s is clicked next frame, it is a new click\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-2, relative=True)
        code = (
            f"else:\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
            f"%(name)s.wasClicked = False  # if %(name)s is clicked next frame, it is a new click\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)

    def writeFrameCodeJS(self, buff):
        BaseVisualComponent.writeFrameCodeJS(self, buff)
        # do writing of init
        inits = getInitVals(self.params, 'PsychoJS')
        # Get callback from params
        callback = inits['callback']
        if inits['callback'].val:
            callback = translatePythonToJavaScript(str(callback))
        else:
            callback = ""

        # Check for current and last button press
        code = (
            "if (%(name)s.status === PsychoJS.Status.STARTED) {\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                "// check whether %(name)s has been pressed\n"
                "if (%(name)s.isClicked) {\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                    "if (!%(name)s.wasClicked) {\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                        "// store time of first click\n"
                        "%(name)s.timesOn.push(%(name)s.clock.getTime());\n"
                        "%(name)s.numClicks += 1;\n"
                        "// store time clicked until\n"
                        "%(name)s.timesOff.push(%(name)s.clock.getTime());\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
                    "} else {\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                    "// update time clicked until;\n"
                    "%(name)s.timesOff[%(name)s.timesOff.length - 1] = %(name)s.clock.getTime();\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
                    "}\n"
        )
        buff.writeIndentedLines(code % inits)

        if self.params['oncePerClick'] or self.params['forceEndRoutine']:
            code = (
                    "if (!%(name)s.wasClicked) {\n"
            )
            buff.writeIndentedLines(code % inits)
            buff.setIndentLevel(1, relative=True)
            if self.params['forceEndRoutine']:
                code = (
                    "// end routine when %(name)s is clicked\n"
                    "continueRoutine = false;\n"
                )
                buff.writeIndentedLines(code % inits)
            if self.params['oncePerClick']:
                buff.writeIndentedLines(callback % inits)
            buff.setIndentLevel(-1, relative=True)
            code = (
                    "}\n"
            )
            buff.writeIndentedLines(code % inits)
        if not self.params['oncePerClick']:
            buff.writeIndentedLines(callback % inits)

        # Store current button press as last
        code = (
                    "// if %(name)s is still clicked next frame, it is not a new click\n"
                    "%(name)s.wasClicked = true;\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
                "} else {\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                    "// if %(name)s is clicked next frame, it is a new click\n"
                    "%(name)s.wasClicked = false;\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
                "}\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            "} else {\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                "// keep clock at 0 if %(name)s hasn't started / has finished\n"
                "%(name)s.clock.reset();\n"
                "// if %(name)s is clicked next frame, it is a new click\n"
                "%(name)s.wasClicked = false;\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            "}\n"
        )
        buff.writeIndentedLines(code % inits)

    def writeRoutineEndCode(self, buff):
        BaseVisualComponent.writeRoutineEndCode(self, buff)
        if len(self.exp.flow._loopList):
            currLoop = self.exp.flow._loopList[-1]  # last (outer-most) loop
        else:
            currLoop = self.exp._expHandler
        name = self.params['name']
        if self.params['save'] == 'first click':
            index = "[0]"
        elif self.params['save'] == 'last click':
            index = "[-1]"
        else:
            index = ""
        if self.params['save'] != 'none':
            code = (
                f"{currLoop.params['name']}.addData('{name}.numClicks', {name}.numClicks)\n"
                f"if {name}.numClicks:\n"
                f"   {currLoop.params['name']}.addData('{name}.timesOn', {name}.timesOn{index})\n"
                f"   {currLoop.params['name']}.addData('{name}.timesOff', {name}.timesOff{index})\n"
                f"else:\n"
                f"   {currLoop.params['name']}.addData('{name}.timesOn', \"\")\n"
                f"   {currLoop.params['name']}.addData('{name}.timesOff', \"\")\n"
            )
            buff.writeIndentedLines(code)

    def writeRoutineEndCodeJS(self, buff):
        # Save data
        code = (
            "psychoJS.experiment.addData('%(name)s.numClicks', %(name)s.numClicks);\n"
            "psychoJS.experiment.addData('%(name)s.timesOn', %(name)s.timesOn);\n"
            "psychoJS.experiment.addData('%(name)s.timesOff', %(name)s.timesOff);\n"
        )
        buff.writeIndentedLines(code % self.params)

    def integrityCheck(self):
        super().integrityCheck()  # run parent class checks first
        alerttools.testFont(self) # Test whether font is available locally


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Simple example of Cetoni neMESYS syringe pump control, based on the `pyqmix`
library. The syringe pump system is described in the following publication:

    CA Andersen, L Alfine, K Ohla, & R Höchenberger (2018):
    "A new gustometer: Template for the construction of a portable and
     modular stimulator for taste and lingual touch."
    Behavior Research Methods. doi: 10.3758/s13428-018-1145-1

"""

from psychopy import event
from psychopy import core
from psychopy.visual import Window, TextStim
from psychopy.hardware import qmix


print('Supported syringe types: %s' % qmix.syringeTypes)
print('Supported volume units: %s' % qmix.volumeUnits)
print('Supported flow rate units: %s' % qmix.flowRateUnits)

# Initialize the first pump (index 0). We assume the pump is
# equipped with a 50 mL glass syringe.
pump = qmix.Pump(index=0,
                 volumeUnit='mL',
                 flowRateUnit='mL/s',
                 syringeType='50 mL glass')

print('Max. flow rate: .3%f %s' % (pump.maxFlowRate, pump.flowRateUnit))

win = Window()
msg = ('Press one of the following keys: \n\n'
       '    F – Fill Syringe at 1 mL/s\n'
       '    E – Empty Syringe at 1 mL/s\n'
       '    A – Aspirate 1 mL at 1 mL/s\n'
       '    D – Dispense 1 mL at 1 mL/s\n'
       '\n'
       '    Q – Quit')
t = TextStim(win, msg)

event.clearEvents()
while True:
    t.draw()
    win.flip()

    # Retrieve keypresses. The user can completely fill or empty the syringe,
    # or aspirate or dispense a small volume (1 mL) by pressing the
    # corresponding keys.
    #
    # When aspirating or dispensing, the code halts further script execution
    # until the pump operation has finished, and then immediately switches the
    # valve position (i.e., from inlet to outlet after aspiration, and from
    # outlet to inlet after dispense). During an experiment, this can ensure
    # a sharp(er) stimulus offset.

    keys = event.getKeys(keyList=['f', 'e', 'a', 'd', 'q'])
    if 'f' in keys:
        pump.fill(flowRate=1, waitUntilDone=False)
    elif 'e' in keys:
        pump.empty(flowRate=1, waitUntilDone=False)
    elif 'a' in keys:
        pump.aspirate(volume=1,
                      flowRate=1,
                      waitUntilDone=True,
                      switchValveWhenDone=True)
    elif 'd' in keys:
        pump.dispense(volume=1,
                      flowRate=1,
                      waitUntilDone=True,
                      switchValveWhenDone=True)
    elif 'q' in keys:
        break

# Immdiately halt all pump operation and shutdown PsychoPy.
pump.stop()
core.quit()



from . import _TestBaseComponentsMixin, _TestDisabledMixin
from psychopy import experiment
import inspect


class _Generic(_TestBaseComponentsMixin, _TestDisabledMixin):
    def __init__(self, compClass):
        self.exp = experiment.Experiment()
        self.rt = experiment.routines.Routine(exp=self.exp, name="testRoutine")
        self.comp = compClass(exp=self.exp, parentName="testRoutine", name=f"test{compClass.__name__}")


def test_all_components():
    for compName, compClass in experiment.getAllComponents().items():
        if compName == "SettingsComponent":
            continue
        # Make a generic testing object for this component
        tester = _Generic(compClass)
        # Run each method from _TestBaseComponentsMixin on tester
        for attr, meth in _TestBaseComponentsMixin.__dict__.items():
            if inspect.ismethod(meth):
                meth(tester)
        # Run each method from _TestBaseComponentsMixin on tester
        for attr, meth in _TestDisabledMixin.__dict__.items():
            if inspect.ismethod(meth):
                meth(tester)
